# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
"""
Push events to Gerrit
"""

from __future__ import annotations

import time
import warnings

from packaging.version import parse as parse_version
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.protocol import ProcessProtocol
from twisted.python import log
from zope.interface import implementer

from buildbot import interfaces
from buildbot.process.results import EXCEPTION
from buildbot.process.results import FAILURE
from buildbot.process.results import RETRY
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.process.results import Results
from buildbot.reporters import utils
from buildbot.reporters.base import ReporterBase
from buildbot.util import bytes2unicode

# Cache the version that the gerrit server is running for this many seconds
GERRIT_VERSION_CACHE_TIMEOUT = 600

GERRIT_LABEL_VERIFIED = 'Verified'
GERRIT_LABEL_REVIEWED = 'Code-Review'


def makeReviewResult(message, *labels):
    """
    helper to produce a review result
    """
    return {"message": message, "labels": dict(labels)}


def _old_add_label(label, value):
    if label == GERRIT_LABEL_VERIFIED:
        return [f"--verified {int(value)}"]
    elif label == GERRIT_LABEL_REVIEWED:
        return [f"--code-review {int(value)}"]
    warnings.warn(
        f'Gerrit older than 2.6 does not support custom labels. Setting {label} is ignored.',
        stacklevel=1,
    )
    return []


def _new_add_label(label, value):
    return [f"--label {label}={int(value)}"]


def defaultReviewCB(builderName, build, result, master, arg):
    if result == RETRY:
        return makeReviewResult(None)

    message = "Buildbot finished compiling your patchset\n"
    message += f"on configuration: {builderName}\n"
    message += f"The result is: {Results[result].upper()}\n"

    return makeReviewResult(message, (GERRIT_LABEL_VERIFIED, result == SUCCESS or -1))


def defaultSummaryCB(buildInfoList, results, master, arg):
    success = False
    failure = False

    msgs = []

    for buildInfo in buildInfoList:
        msg = f"Builder {buildInfo['name']} {buildInfo['resultText']} ({buildInfo['text']})"
        link = buildInfo.get('url', None)
        if link:
            msg += " - " + link
        else:
            msg += "."
        msgs.append(msg)

        if buildInfo['result'] == SUCCESS:  # pylint: disable=simplifiable-if-statement
            success = True
        else:
            failure = True

    if success and not failure:
        verified = 1
    else:
        verified = -1

    return makeReviewResult('\n\n'.join(msgs), (GERRIT_LABEL_VERIFIED, verified))


# These are just sentinel values for GerritStatusPush.__init__ args
class DEFAULT_REVIEW:
    pass


class DEFAULT_SUMMARY:
    pass


@defer.inlineCallbacks
def extract_project_revision(master, report):
    props = None
    if report["builds"]:
        props = report["builds"][0].get("properties", None)

    if props is None:
        props = yield master.data.get(("buildsets", report["buildset"]["bsid"], "properties"))

    def get_property(props, name):
        if props is None:
            return None
        return props.get(name, [None])[0]

    # Gerrit + Repo
    downloads = get_property(props, "repo_downloads")
    downloaded = get_property(props, "repo_downloaded")
    if downloads is not None and downloaded is not None:
        downloaded = downloaded.split(" ")
        if downloads and 2 * len(downloads) == len(downloaded):
            for i, download in enumerate(downloads):
                try:
                    project, change1 = download.split(" ")
                except ValueError:
                    return None, None  # something is wrong, abort
                change2 = downloaded[2 * i]
                revision = downloaded[2 * i + 1]
                if change1 == change2:
                    return project, revision
                else:
                    return None, None
        return None, None

    # Gerrit + Git
    # used only to verify Gerrit source
    if get_property(props, "event.change.id") is not None:
        project = get_property(props, "event.change.project")
        codebase = get_property(props, "codebase")
        revision = (
            get_property(props, "event.patchSet.revision")
            or get_property(props, "got_revision")
            or get_property(props, "revision")
        )

        if isinstance(revision, dict):
            # in case of the revision is a codebase revision, we just take
            # the revisionfor current codebase
            if codebase is not None:
                revision = revision[codebase]
            else:
                revision = None

        return project, revision

    return None, None


class GerritStatusGeneratorBase:
    def __init__(self, callback, callback_arg, builders, want_steps, want_logs):
        self.callback = callback
        self.callback_arg = callback_arg
        self.builders = builders
        self.want_steps = want_steps
        self.want_logs = want_logs

    def is_build_reported(self, build):
        return self.builders is None or build["builder"]["name"] in self.builders

    @defer.inlineCallbacks
    def get_build_details(self, master, build):
        br = yield master.data.get(("buildrequests", build["buildrequestid"]))
        buildset = yield master.data.get(("buildsets", br["buildsetid"]))
        yield utils.getDetailsForBuilds(
            master, buildset, [build], want_properties=True, want_steps=self.want_steps
        )


@implementer(interfaces.IReportGenerator)
class GerritBuildSetStatusGenerator(GerritStatusGeneratorBase):
    wanted_event_keys = [
        ("buildsets", None, "complete"),
    ]

    def check(self):
        pass

    @defer.inlineCallbacks
    def generate(self, master, reporter, key, message):
        bsid = message["bsid"]
        res = yield utils.getDetailsForBuildset(
            master,
            bsid,
            want_properties=True,
            want_steps=self.want_steps,
            want_logs=self.want_logs,
            want_logs_content=self.want_logs,
        )

        builds = res["builds"]
        buildset = res["buildset"]

        builds = [build for build in builds if self.is_build_reported(build)]
        if not builds:
            return None

        def get_build_info(build):
            result = build["results"]
            resultText = {
                SUCCESS: "succeeded",
                FAILURE: "failed",
                WARNINGS: "completed with warnings",
                EXCEPTION: "encountered an exception",
            }.get(result, f"completed with unknown result {result}")

            return {
                "name": build["builder"]["name"],
                "result": result,
                "resultText": resultText,
                "text": build["state_string"],
                "url": utils.getURLForBuild(master, build["builder"]["builderid"], build["number"]),
                "build": build,
            }

        build_info_list = sorted(
            [get_build_info(build) for build in builds], key=lambda bi: bi["name"]
        )

        result = yield self.callback(
            build_info_list, Results[buildset["results"]], master, self.callback_arg
        )

        return {
            "body": result.get("message", None),
            "extra_info": {
                "labels": result.get("labels"),
            },
            "builds": [builds[0]],
            "buildset": buildset,
        }


@implementer(interfaces.IReportGenerator)
class GerritBuildStartStatusGenerator(GerritStatusGeneratorBase):
    wanted_event_keys = [
        ("builds", None, "new"),
    ]

    def check(self):
        pass

    @defer.inlineCallbacks
    def generate(self, master, reporter, key, message):
        build = message
        yield self.get_build_details(master, build)
        if not self.is_build_reported(build):
            return None

        result = yield self.callback(build["builder"]["name"], build, self.callback_arg)

        return {
            "body": result.get("message", None),
            "extra_info": {
                "labels": result.get("labels"),
            },
            "builds": [build],
            "buildset": build["buildset"],
        }


@implementer(interfaces.IReportGenerator)
class GerritBuildEndStatusGenerator(GerritStatusGeneratorBase):
    wanted_event_keys = [
        ('builds', None, 'finished'),
    ]

    def check(self):
        pass

    @defer.inlineCallbacks
    def generate(self, master, reporter, key, message):
        build = message
        yield self.get_build_details(master, build)
        if not self.is_build_reported(build):
            return None

        result = yield self.callback(
            build['builder']['name'], build, build['results'], master, self.callback_arg
        )

        return {
            "body": result.get("message", None),
            "extra_info": {
                "labels": result.get("labels"),
            },
            "builds": [build],
            "buildset": build["buildset"],
        }


class GerritStatusPush(ReporterBase):
    """Event streamer to a gerrit ssh server."""

    name: str | None = "GerritStatusPush"  # type: ignore[assignment]
    gerrit_server = None
    gerrit_username = None
    gerrit_port = None
    gerrit_version_time = None
    gerrit_version = None
    gerrit_identity_file = None
    _gerrit_notify = None

    def checkConfig(
        self,
        server,
        username,
        port=29418,
        identity_file=None,
        notify=None,
        generators=None,
        **kwargs,
    ):
        if generators is None:
            generators = []
            generators.append(
                GerritBuildSetStatusGenerator(
                    callback=defaultSummaryCB,
                    callback_arg=None,
                    builders=None,
                    want_steps=False,
                    want_logs=False,
                )
            )

        super().checkConfig(generators=generators, **kwargs)

    def reconfigService(
        self,
        server,
        username,
        port=29418,
        identity_file=None,
        notify=None,
        generators=None,
        **kwargs,
    ):
        self.gerrit_server = server
        self.gerrit_username = username
        self.gerrit_port = port
        self.gerrit_version = None
        self.gerrit_version_time = 0
        self.gerrit_identity_file = identity_file
        self._gerrit_notify = notify

        if generators is None:
            generators = []
            generators.append(
                GerritBuildSetStatusGenerator(
                    callback=defaultSummaryCB,
                    callback_arg=None,
                    builders=None,
                    want_steps=False,
                    want_logs=False,
                )
            )

        super().reconfigService(generators=generators, **kwargs)

    def _gerritCmd(self, *args):
        """Construct a command as a list of strings suitable for
        :func:`subprocess.call`.
        """
        if self.gerrit_identity_file is not None:
            options = ['-i', self.gerrit_identity_file]
        else:
            options = []
        return [
            'ssh',
            '-o',
            'BatchMode=yes',
            *options,
            '@'.join((self.gerrit_username, self.gerrit_server)),
            '-p',
            str(self.gerrit_port),
            'gerrit',
            *list(args),
        ]

    class VersionPP(ProcessProtocol):
        def __init__(self, func):
            self.func = func
            self.gerrit_version = None

        def outReceived(self, data):
            vstr = b"gerrit version "
            if not data.startswith(vstr):
                log.msg(b"Error: Cannot interpret gerrit version info: " + data)
                return
            vers = data[len(vstr) :].strip()
            log.msg(b"gerrit version: " + vers)
            self.gerrit_version = parse_version(bytes2unicode(vers))

        def errReceived(self, data):
            log.msg(b"gerriterr: " + data)

        def processEnded(self, reason):
            if reason.value.exitCode:
                log.msg("gerrit version status: ERROR:", reason)
                return
            if self.gerrit_version:
                self.func(self.gerrit_version)

    def getCachedVersion(self):
        if self.gerrit_version is None:
            return None
        if time.time() - self.gerrit_version_time > GERRIT_VERSION_CACHE_TIMEOUT:
            # cached version has expired
            self.gerrit_version = None
        return self.gerrit_version

    def processVersion(self, gerrit_version, func):
        self.gerrit_version = gerrit_version
        self.gerrit_version_time = time.time()
        func()

    def callWithVersion(self, func):
        command = self._gerritCmd("version")

        def callback(gerrit_version):
            return self.processVersion(gerrit_version, func)

        self.spawnProcess(self.VersionPP(callback), command[0], command, env=None)

    class LocalPP(ProcessProtocol):
        def __init__(self, status):
            self.status = status

        def outReceived(self, data):
            log.msg("gerritout:", data)

        def errReceived(self, data):
            log.msg("gerriterr:", data)

        def processEnded(self, reason):
            if reason.value.exitCode:
                log.msg("gerrit status: ERROR:", reason)
            else:
                log.msg("gerrit status: OK")

    @defer.inlineCallbacks
    def sendMessage(self, reports):
        report = reports[0]

        project, revision = yield extract_project_revision(self.master, report)

        if report["body"] is None or project is None or revision is None:
            return None

        labels = None
        extra_info = report.get("extra_info", None)
        if extra_info is not None:
            labels = extra_info.get("labels", None)

        if labels is None and report.get("builds", None):
            # At least one build
            success = False
            failure = False
            pending = False

            for build in report["builds"]:
                if build["results"] is None:
                    pending = True
                elif build["results"] == SUCCESS:
                    success = True
                else:
                    failure = True

            if failure:
                verified = -1
            elif pending:
                verified = 0
            elif success:
                verified = 1
            else:
                verified = -1

            labels = {GERRIT_LABEL_VERIFIED: verified}

        self.send_code_review(project, revision, report["body"], labels)
        return None

    def send_code_review(self, project, revision, message, labels):
        gerrit_version = self.getCachedVersion()
        if gerrit_version is None:
            self.callWithVersion(lambda: self.send_code_review(project, revision, message, labels))
            return

        assert gerrit_version
        command = self._gerritCmd("review", f"--project {project}")

        if gerrit_version >= parse_version("2.13"):
            command.append('--tag autogenerated:buildbot')

        if self._gerrit_notify is not None:
            command.append(f'--notify {self._gerrit_notify!s}')

        if message:
            message = message.replace("'", "\"")
            command.append(f"--message '{message}'")

        if labels:
            if gerrit_version < parse_version("2.6"):
                add_label = _old_add_label
            else:
                add_label = _new_add_label

            for label, value in labels.items():
                command.extend(add_label(label, value))

        command.append(revision)
        command = [str(s) for s in command]
        self.spawnProcess(self.LocalPP(self), command[0], command, env=None)

    def spawnProcess(self, *arg, **kw):
        reactor.spawnProcess(*arg, **kw)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot import config as config_module
from buildbot.master import BuildMaster
from buildbot.scripts import base
from buildbot.util import in_reactor

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection


async def doCleanupDatabase(config, master_cfg) -> None:
    if not config['quiet']:
        print(f"cleaning database ({master_cfg.db.db_url})")

    master = BuildMaster(config['basedir'])
    master.config = master_cfg
    db = master.db
    try:
        await db.setup(check_version=False, verbose=not config['quiet'])
        res = await db.logs.getLogs()
        percent = 0
        saved = 0
        for i, log in enumerate(res, start=1):
            saved += await db.logs.compressLog(log.id, force=config['force'])
            if not config['quiet'] and percent != int(i * 100 / len(res)):
                percent = int(i * 100 / len(res))
                print(f" {percent}%  {saved} saved", flush=True)
                saved = 0

        assert master.db._engine is not None
        vacuum_stmt = {
            # https://www.postgresql.org/docs/current/sql-vacuum.html
            'postgresql': f'VACUUM FULL {master.db.model.logchunks.name};',
            # https://dev.mysql.com/doc/refman/5.7/en/optimize-table.html
            'mysql': f'OPTIMIZE TABLE {master.db.model.logchunks.name};',
            # https://www.sqlite.org/lang_vacuum.html
            'sqlite': 'vacuum;',
        }.get(master.db._engine.dialect.name)

        if vacuum_stmt is not None:

            def thd(conn: Connection) -> None:
                if not config['quiet']:
                    print(f"executing vacuum operation '{vacuum_stmt}'...", flush=True)

                # vacuum operation cannot be done in a transaction
                # https://github.com/sqlalchemy/sqlalchemy/discussions/6959#discussioncomment-1251681
                with conn.execution_options(isolation_level='AUTOCOMMIT'):
                    conn.exec_driver_sql(vacuum_stmt).close()

                conn.commit()

            await db.pool.do(thd)
    finally:
        await db.pool.stop()


@in_reactor
async def cleanupDatabase(config):  # pragma: no cover
    # we separate the actual implementation to protect unit tests
    # from @in_reactor which stops the reactor
    return defer.Deferred.fromCoroutine(_cleanupDatabase(config))


async def _cleanupDatabase(config) -> int:
    if not base.checkBasedir(config):
        return 1

    config['basedir'] = os.path.abspath(config['basedir'])

    orig_cwd = os.getcwd()

    try:
        os.chdir(config['basedir'])

        with base.captureErrors(
            (SyntaxError, ImportError), f"Unable to load 'buildbot.tac' from '{config['basedir']}':"
        ):
            configFile = base.getConfigFileFromTac(config['basedir'])

        with base.captureErrors(
            config_module.ConfigErrors, f"Unable to load '{configFile}' from '{config['basedir']}':"
        ):
            master_cfg = base.loadConfig(config, configFile)

        if not master_cfg:
            return 1

        await doCleanupDatabase(config, master_cfg)

        if not config['quiet']:
            print("cleanup complete")
    finally:
        os.chdir(orig_cwd)

    return 0



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import Callable

import zope.interface.interface
from zope.interface.interface import Attribute


class InterfaceTests:
    # assertions
    assertEqual: Callable[..., None]

    def assertArgSpecMatches(self, actualMethod, *fakeMethods):
        """Usage::

            @self.assertArgSpecMatches(obj.methodUnderTest)
            def methodTemplate(self, arg1, arg2):
                pass

        or, more useful when you will be faking out C{methodUnderTest}:

            self.assertArgSpecMatches(obj.methodUnderTest, self.fakeMethod)
        """

        def filter(signature: inspect.Signature):
            parameters = OrderedDict(signature.parameters)
            for name in parameters:
                if name == 'self':
                    parameters.pop('self')
                break

            delete_names = []
            for name in parameters:
                if name.startswith('_'):
                    delete_names.append(name)
            for name in delete_names:
                parameters.pop(name)

            # Remove all type annotations
            # as they can be stored as str when quoted or when `__future__.annotations`
            # is imported, we can't check whether the types are compatible.
            # Type checking should be left to a static type checker
            signature = signature.replace(return_annotation=inspect.Signature.empty)
            for name, param in parameters.items():
                parameters[name] = param.replace(annotation=inspect.Parameter.empty)

            signature = signature.replace(parameters=list(parameters.values()))
            return signature

        def remove_decorators(func):
            try:
                return func.__wrapped__
            except AttributeError:
                return func

        def filter_argspec(func):
            return filter(inspect.signature(remove_decorators(func)))

        def assert_same_argspec(expected, actual):
            if expected != actual:
                msg = f"Expected: {expected}; got: {actual}"
                self.fail(msg)

        actual_argspec = filter_argspec(actualMethod)

        for fakeMethod in fakeMethods:
            fake_argspec = filter_argspec(fakeMethod)
            assert_same_argspec(actual_argspec, fake_argspec)

        def assert_same_argspec_decorator(decorated):
            expected_argspec = filter_argspec(decorated)
            assert_same_argspec(expected_argspec, actual_argspec)
            # The decorated function works as usual.
            return decorated

        return assert_same_argspec_decorator

    def assertInterfacesImplemented(self, cls):
        "Given a class, assert that the zope.interface.Interfaces are implemented to specification."

        for interface in zope.interface.implementedBy(cls):
            for attr, template_argspec in interface.namesAndDescriptions():
                if not hasattr(cls, attr):
                    msg = f"Expected: {cls!r}; to implement: {attr} as specified in {interface!r}"
                    self.fail(msg)
                actual_argspec = getattr(cls, attr)
                if isinstance(template_argspec, Attribute):
                    continue
                # else check method signatures
                while hasattr(actual_argspec, '__wrapped__'):
                    actual_argspec = actual_argspec.__wrapped__
                actual_argspec = zope.interface.interface.fromMethod(actual_argspec)

                if actual_argspec.getSignatureInfo() != template_argspec.getSignatureInfo():
                    msg = (
                        f"{attr}: expected: {template_argspec.getSignatureString()}; got: "
                        f"{actual_argspec.getSignatureString()}"
                    )
                    self.fail(msg)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Parse various kinds of 'CVS notify' email.
"""

from __future__ import annotations

import calendar
import datetime
import re
import time
from email import message_from_file
from email.iterators import body_line_iterator
from email.utils import mktime_tz
from email.utils import parseaddr
from email.utils import parsedate_tz
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python import log
from zope.interface import implementer

from buildbot import util
from buildbot.interfaces import IChangeSource
from buildbot.util.maildir import MaildirService

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType


@implementer(IChangeSource)
class MaildirSource(MaildirService, util.ComparableMixin):
    """Generic base class for Maildir-based change sources"""

    compare_attrs: ClassVar[Sequence[str]] = ("basedir", "pollInterval", "prefix")
    # twisted is marked as typed, but doesn't specify this type correctly
    name: str | None = 'MaildirSource'  # type: ignore[assignment]

    def __init__(
        self, maildir: str, prefix: str | None = None, category: str = '', repository: str = ''
    ) -> None:
        super().__init__(maildir)
        self.prefix = prefix
        self.category = category
        self.repository = repository
        if prefix and not prefix.endswith("/"):
            log.msg(
                f"MaildirSource: you probably want your prefix=('{prefix}') to end with a slash"
            )

    def describe(self) -> str:
        return f"{self.__class__.__name__} watching maildir '{self.basedir}'"

    @defer.inlineCallbacks
    def messageReceived(self, filename: str) -> InlineCallbacksType:
        with self.moveToCurDir(filename) as f:
            chtuple = self.parse_file(f, self.prefix)

        src = None
        chdict = None
        if chtuple:
            src, chdict = chtuple
        if chdict:
            yield self.master.data.updates.addChange(src=str(src), **chdict)
        else:
            log.msg(f"no change found in maildir file '{filename}'")

    def parse_file(self, fd: Any, prefix: str | None = None) -> tuple[str, dict[str, Any]] | None:
        m = message_from_file(fd)
        return self.parse(m, prefix)

    def parse(self, m: Any, prefix: str | None = None) -> tuple[str, dict[str, Any]] | None:
        raise NotImplementedError


class CVSMaildirSource(MaildirSource):
    name = "CVSMaildirSource"

    def __init__(
        self,
        maildir: str,
        prefix: str | None = None,
        category: str = '',
        repository: str = '',
        properties: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(maildir, prefix, category, repository)
        if properties is None:
            properties = {}
        self.properties = properties

    def parse(self, m: Any, prefix: str | None = None) -> tuple[str, dict[str, Any]] | None:
        """Parse messages sent by the 'buildbot-cvs-mail' program."""
        # The mail is sent from the person doing the checkin. Assume that the
        # local username is enough to identify them (this assumes a one-server
        # cvs-over-rsh environment rather than the server-dirs-shared-over-NFS
        # model)
        _, addr = parseaddr(m["from"])
        if not addr:
            # no From means this message isn't from buildbot-cvs-mail
            return None
        at = addr.find("@")
        if at == -1:
            author = addr  # might still be useful
        else:
            author = addr[:at]
        author = util.bytes2unicode(author, encoding="ascii")

        # CVS accepts RFC822 dates. buildbot-cvs-mail adds the date as
        # part of the mail header, so use that.
        # This assumes cvs is being access via ssh or pserver, so the time
        # will be the CVS server's time.

        # calculate a "revision" based on that timestamp, or the current time
        # if we're unable to parse the date.
        log.msg('Processing CVS mail')
        dateTuple = parsedate_tz(m["date"])
        if dateTuple is None:
            when = util.now()
        else:
            when = mktime_tz(dateTuple)

        theTime = datetime.datetime.fromtimestamp(float(when), datetime.timezone.utc)
        rev = theTime.strftime('%Y-%m-%d %H:%M:%S')

        catRE = re.compile(r'^Category:\s*(\S.*)')
        cvsRE = re.compile(r'^CVSROOT:\s*(\S.*)')
        cvsmodeRE = re.compile(r'^Cvsmode:\s*(\S.*)')
        filesRE = re.compile(r'^Files:\s*(\S.*)')
        modRE = re.compile(r'^Module:\s*(\S.*)')
        pathRE = re.compile(r'^Path:\s*(\S.*)')
        projRE = re.compile(r'^Project:\s*(\S.*)')
        singleFileRE = re.compile(r'(.*) (NONE|\d(\.|\d)+) (NONE|\d(\.|\d)+)')
        tagRE = re.compile(r'^\s+Tag:\s*(\S.*)')
        updateRE = re.compile(r'^Update of:\s*(\S.*)')
        comments = ""
        branch = None
        category = None
        cvsroot = None
        cvsmode = None
        fileList = None
        files = []
        isdir = 0
        path = None
        project = None

        lines = list(body_line_iterator(m))
        while lines:
            line = lines.pop(0)
            m = catRE.match(line)
            if m:
                category = m.group(1)
                continue
            m = cvsRE.match(line)
            if m:
                cvsroot = m.group(1)
                continue
            m = cvsmodeRE.match(line)
            if m:
                cvsmode = m.group(1)
                continue
            m = filesRE.match(line)
            if m:
                fileList = m.group(1)
                continue
            m = modRE.match(line)
            if m:
                # We don't actually use this
                # module = m.group(1)
                continue
            m = pathRE.match(line)
            if m:
                path = m.group(1)
                continue
            m = projRE.match(line)
            if m:
                project = m.group(1)
                continue
            m = tagRE.match(line)
            if m:
                branch = m.group(1)
                continue
            m = updateRE.match(line)
            if m:
                # We don't actually use this
                # updateof = m.group(1)
                continue
            if line == "Log Message:\n":
                break

        # CVS 1.11 lists files as:
        #   repo/path file,old-version,new-version file2,old-version,new-version
        # Version 1.12 lists files as:
        #   file1 old-version new-version file2 old-version new-version
        #
        # files consists of tuples of 'file-name old-version new-version'
        # The versions are either dotted-decimal version numbers, ie 1.1
        # or NONE. New files are of the form 'NONE NUMBER', while removed
        # files are 'NUMBER NONE'. 'NONE' is a literal string
        # Parsing this instead of files list in 'Added File:' etc
        # makes it possible to handle files with embedded spaces, though
        # it could fail if the filename was 'bad 1.1 1.2'
        # For cvs version 1.11, we expect
        #  my_module new_file.c,NONE,1.1
        #  my_module removed.txt,1.2,NONE
        #  my_module modified_file.c,1.1,1.2
        # While cvs version 1.12 gives us
        #  new_file.c NONE 1.1
        #  removed.txt 1.2 NONE
        #  modified_file.c 1.1,1.2

        if fileList is None:
            log.msg('CVSMaildirSource Mail with no files. Ignoring')
            return None  # We don't have any files. Email not from CVS

        if cvsmode == '1.11':
            # Please, no repo paths with spaces!
            m = re.search('([^ ]*) ', fileList)
            if m:
                path = m.group(1)
            else:
                log.msg('CVSMaildirSource can\'t get path from file list. Ignoring mail')
                return None
            fileList = fileList[len(path) :].strip()
            singleFileRE = re.compile(
                r'(.+?),(NONE|(?:\d+\.(?:\d+\.\d+\.)*\d+)),(NONE|(?:\d+\.(?:\d+\.\d+\.)*\d+))(?: |$)'
            )
        elif cvsmode == '1.12':
            singleFileRE = re.compile(
                r'(.+?) (NONE|(?:\d+\.(?:\d+\.\d+\.)*\d+)) (NONE|(?:\d+\.(?:\d+\.\d+\.)*\d+))(?: |$)'
            )
            if path is None:
                raise ValueError('CVSMaildirSource cvs 1.12 require path. Check cvs loginfo config')
        else:
            raise ValueError(f'Expected cvsmode 1.11 or 1.12. got: {cvsmode}')

        log.msg(f"CVSMaildirSource processing filelist: {fileList}")
        while fileList:
            m = singleFileRE.match(fileList)
            if m:
                curFile = path + '/' + m.group(1)
                files.append(curFile)
                fileList = fileList[m.end() :]
            else:
                log.msg('CVSMaildirSource no files matched regex. Ignoring')
                return None  # bail - we couldn't parse the files that changed
        # Now get comments
        while lines:
            line = lines.pop(0)
            comments += line

        comments = comments.rstrip() + "\n"

        return (
            'cvs',
            {
                "author": author,
                "committer": None,
                "files": files,
                "comments": None if comments == "\n" else comments,
                "isdir": isdir,
                "when": when,
                "branch": branch,
                "revision": rev,
                "category": category,
                "repository": cvsroot,
                "project": project,
                "properties": self.properties,
            },
        )


# svn "commit-email.pl" handler.  The format is very similar to freshcvs mail;
# here's a sample:

#  From: username [at] apache.org    [slightly obfuscated to avoid spam here]
#  To: commits [at] spamassassin.apache.org
#  Subject: svn commit: r105955 - in spamassassin/trunk: . lib/Mail
#  ...
#
#  Author: username
#  Date: Sat Nov 20 00:17:49 2004      [note: TZ = local tz on server!]
#  New Revision: 105955
#
#  Modified:   [also Removed: and Added:]
#    [filename]
#    ...
#  Log:
#  [log message]
#  ...
#
#
#  Modified: spamassassin/trunk/lib/Mail/SpamAssassin.pm
#  [unified diff]
#
#  [end of mail]


class SVNCommitEmailMaildirSource(MaildirSource):
    name = "SVN commit-email.pl"

    def parse(self, m: Any, prefix: str | None = None) -> tuple[str, dict[str, Any]] | None:
        """Parse messages sent by the svn 'commit-email.pl' trigger."""

        # The mail is sent from the person doing the checkin. Assume that the
        # local username is enough to identify them (this assumes a one-server
        # cvs-over-rsh environment rather than the server-dirs-shared-over-NFS
        # model)
        _, addr = parseaddr(m["from"])
        if not addr:
            return None  # no From means this message isn't from svn
        at = addr.find("@")
        if at == -1:
            author = addr  # might still be useful
        else:
            author = addr[:at]

        # we take the time of receipt as the time of checkin. Not correct (it
        # depends upon the email latency), but it avoids the
        # out-of-order-changes issue. Also syncmail doesn't give us anything
        # better to work with, unless you count pulling the v1-vs-v2
        # timestamp out of the diffs, which would be ugly. TODO: Pulling the
        # 'Date:' header from the mail is a possibility, and
        # email.utils.parsedate_tz may be useful. It should be configurable,
        # however, because there are a lot of broken clocks out there.
        when = util.now()

        files = []
        comments = ""
        lines = list(body_line_iterator(m))
        rev = None
        while lines:
            line = lines.pop(0)

            # "Author: jmason"
            match = re.search(r"^Author: (\S+)", line)
            if match:
                author = match.group(1)

            # "New Revision: 105955"
            match = re.search(r"^New Revision: (\d+)", line)
            if match:
                rev = match.group(1)

            # possible TODO: use "Date: ..." data here instead of time of
            # commit message receipt, above. however, this timestamp is
            # specified *without* a timezone, in the server's local TZ, so to
            # be accurate buildbot would need a config setting to specify the
            # source server's expected TZ setting! messy.

            # this stanza ends with the "Log:"
            if line == "Log:\n":
                break

        # commit message is terminated by the file-listing section
        while lines:
            line = lines.pop(0)
            if line in ("Modified:\n", "Added:\n", "Removed:\n"):
                break
            comments += line
        comments = comments.rstrip() + "\n"

        while lines:
            line = lines.pop(0)
            if line == "\n":
                break
            if line.find("Modified:\n") == 0:
                continue  # ignore this line
            if line.find("Added:\n") == 0:
                continue  # ignore this line
            if line.find("Removed:\n") == 0:
                continue  # ignore this line
            line = line.strip()

            thesefiles = line.split(" ")
            for f in thesefiles:
                if prefix:
                    # insist that the file start with the prefix: we may get
                    # changes we don't care about too
                    if f.startswith(prefix):
                        f = f[len(prefix) :]
                    else:
                        log.msg(
                            f"ignored file from svn commit: prefix '{prefix}' "
                            f"does not match filename '{f}'"
                        )
                        continue

                # TODO: figure out how new directories are described, set
                # .isdir
                files.append(f)

        if not files:
            log.msg("no matching files found, ignoring commit")
            return None

        return (
            'svn',
            {
                "author": author,
                "committer": None,
                "files": files,
                "comments": comments,
                "when": when,
                "revision": rev,
            },
        )


# bzr Launchpad branch subscription mails. Sample mail:
#
#   From: noreply@launchpad.net
#   Subject: [Branch ~knielsen/maria/tmp-buildbot-test] Rev 2701: test add file
#   To: Joe <joe@acme.com>
#   ...
#
#   ------------------------------------------------------------
#   revno: 2701
#   committer: Joe <joe@acme.com>
#   branch nick: tmpbb
#   timestamp: Fri 2009-05-15 10:35:43 +0200
#   message:
#     test add file
#   added:
#     test-add-file
#
#
#   --
#
#   https://code.launchpad.net/~knielsen/maria/tmp-buildbot-test
#
#   You are subscribed to branch lp:~knielsen/maria/tmp-buildbot-test.
#   To unsubscribe from this branch go to
#   https://code.launchpad.net/~knielsen/maria/tmp-buildbot-test/+edit-subscription.
#
# [end of mail]


class BzrLaunchpadEmailMaildirSource(MaildirSource):
    name = "Launchpad"

    compare_attrs: ClassVar[Sequence[str]] = ("branchMap", "defaultBranch")

    def __init__(
        self,
        maildir: str,
        prefix: str | None = None,
        branchMap: dict[str, str] | None = None,
        defaultBranch: str | None = None,
        **kwargs: Any,
    ) -> None:
        self.branchMap = branchMap
        self.defaultBranch = defaultBranch
        super().__init__(maildir, prefix, **kwargs)

    def parse(self, m: Any, prefix: str | None = None) -> tuple[str, dict[str, Any]] | None:
        """Parse branch notification messages sent by Launchpad."""

        subject = m["subject"]
        match = re.search(r"^\s*\[Branch\s+([^]]+)\]", subject)
        if match:
            repository = match.group(1)
        else:
            repository = None

        # Put these into a dictionary, otherwise we cannot assign them
        # from nested function definitions.
        d: dict[str, Any] = {'files': [], 'comments': ""}
        gobbler = None
        rev = None
        author = None
        when = util.now()

        def gobble_comment(s: str) -> None:
            d['comments'] += s + "\n"

        def gobble_removed(s: str) -> None:
            d['files'].append(f'{s} REMOVED')

        def gobble_added(s: str) -> None:
            d['files'].append(f'{s} ADDED')

        def gobble_modified(s: str) -> None:
            d['files'].append(f'{s} MODIFIED')

        def gobble_renamed(s: str) -> None:
            match = re.search(r"^(.+) => (.+)$", s)
            if match:
                d['files'].append(f'{match.group(1)} RENAMED {match.group(2)}')
            else:
                d['files'].append(f'{s} RENAMED')

        lines = list(body_line_iterator(m, True))
        rev = None
        while lines:
            line = str(lines.pop(0), encoding="utf-8", errors="ignore")  # type: ignore[call-overload]

            # revno: 101
            match = re.search(r"^revno: ([0-9.]+)", line)
            if match:
                rev = match.group(1)

            # committer: Joe <joe@acme.com>
            match = re.search(r"^committer: (.*)$", line)
            if match:
                author = match.group(1)

            # timestamp: Fri 2009-05-15 10:35:43 +0200
            # datetime.strptime() is supposed to support %z for time zone, but
            # it does not seem to work. So handle the time zone manually.
            match = re.search(
                r"^timestamp: [a-zA-Z]{3} (\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}) ([-+])(\d{2})(\d{2})$",
                line,
            )
            if match:
                datestr = match.group(1)
                tz_sign = match.group(2)
                tz_hours = match.group(3)
                tz_minutes = match.group(4)
                when = parseLaunchpadDate(datestr, tz_sign, tz_hours, tz_minutes)

            if re.search(r"^message:\s*$", line):
                gobbler = gobble_comment
            elif re.search(r"^removed:\s*$", line):
                gobbler = gobble_removed
            elif re.search(r"^added:\s*$", line):
                gobbler = gobble_added
            elif re.search(r"^renamed:\s*$", line):
                gobbler = gobble_renamed
            elif re.search(r"^modified:\s*$", line):
                gobbler = gobble_modified
            elif re.search(r"^  ", line) and gobbler:
                gobbler(line[2:-1])  # Use :-1 to gobble trailing newline

        # Determine the name of the branch.
        branch = None
        if self.branchMap and repository:
            if repository in self.branchMap:
                branch = self.branchMap[repository]
            elif "lp:" + repository in self.branchMap:
                branch = self.branchMap['lp:' + repository]
        if not branch:
            if self.defaultBranch:
                branch = self.defaultBranch
            else:
                if repository:
                    branch = 'lp:' + repository
                else:
                    branch = None

        if rev and author:
            return (
                'bzr',
                {
                    "author": author,
                    "committer": None,
                    "files": d['files'],
                    "comments": d['comments'],
                    "when": when,
                    "revision": rev,
                    "branch": branch,
                    "repository": repository or '',
                },
            )
        return None


def parseLaunchpadDate(datestr: str, tz_sign: str, tz_hours: str, tz_minutes: str) -> float:
    time_no_tz = calendar.timegm(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    tz_delta = 60 * 60 * int(tz_sign + tz_hours) + 60 * int(tz_minutes)
    return time_no_tz - tz_delta



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python import log

from buildbot import util
from buildbot.util import service
from buildbot.util import subscription
from buildbot.util.eventual import eventually

if False:  # for debugging  pylint: disable=using-constant-test
    debuglog = log.msg
else:
    debuglog = lambda m: None


class BaseLock:
    """
    Class handling claiming and releasing of L{self}, and keeping track of
    current and waiting owners.

    We maintain the wait queue in FIFO order, and ensure that counting waiters
    in the queue behind exclusive waiters cannot acquire the lock. This ensures
    that exclusive waiters are not starved.
    """

    description = "<BaseLock>"

    def __init__(self, name, maxCount=1):
        super().__init__()

        # Name of the lock
        self.lockName = name
        # Current queue, tuples (waiter_id, LockAccess, deferred)
        self.waiting = []
        # Current owners, tuples (owner_id, LockAccess)
        self.owners = []
        # maximal number of counting owners
        self.maxCount = maxCount

        # current number of claimed exclusive locks (0 or 1), must match
        # self.owners
        self._claimed_excl = 0

        # current number of claimed counting locks (0 to self.maxCount), must
        # match self.owners. Note that self.maxCount is not a strict limit, the
        # number of claimed counting locks may be higher than self.maxCount if
        # it was lowered by
        self._claimed_counting = 0

        # subscriptions to this lock being released
        self.release_subs = subscription.SubscriptionPoint(f"{self!r} releases")

    def __repr__(self):
        return self.description

    def setMaxCount(self, count):
        old_max_count = self.maxCount
        self.maxCount = count

        if count > old_max_count:
            self._tryWakeUp()

    def _find_waiting(self, requester):
        for idx, waiter in enumerate(self.waiting):
            if waiter[0] == id(requester):
                return idx
        return None

    def isAvailable(self, requester, access):
        """Return a boolean whether the lock is available for claiming"""
        debuglog(f"{self} isAvailable({requester}, {access}): self.owners={self.owners!r}")
        num_excl = self._claimed_excl
        num_counting = self._claimed_counting

        if not access.count:
            return True

        w_index = self._find_waiting(requester)
        if w_index is None:
            w_index = len(self.waiting)

        ahead = self.waiting[:w_index]

        if access.mode == 'counting':
            # Wants counting access
            return (
                not num_excl
                and num_counting + len(ahead) + access.count <= self.maxCount
                and all(w[1].mode == 'counting' for w in ahead)
            )
        # else Wants exclusive access
        return not num_excl and not num_counting and not ahead

    def _addOwner(self, owner, access):
        self.owners.append((id(owner), access))
        if access.mode == 'counting':
            self._claimed_counting += access.count
        else:
            self._claimed_excl += 1

        assert (self._claimed_excl and not self._claimed_counting) or (
            not self._claimed_excl and self._claimed_excl <= self.maxCount
        )

    def _removeOwner(self, owner, access):
        # returns True if owner removed, False if the lock has been already
        # released
        entry = (id(owner), access)
        if entry not in self.owners:
            return False

        self.owners.remove(entry)
        if access.mode == 'counting':
            self._claimed_counting -= access.count
        else:
            self._claimed_excl -= 1
        return True

    def claim(self, owner, access):
        """Claim the lock (lock must be available)"""
        debuglog(f"{self} claim({owner}, {access.mode})")
        assert owner is not None
        assert self.isAvailable(owner, access), "ask for isAvailable() first"

        assert isinstance(access, LockAccess)
        assert access.mode in ['counting', 'exclusive']
        assert isinstance(access.count, int)
        if access.mode == 'exclusive':
            assert access.count == 1
        else:
            assert access.count >= 0
        if not access.count:
            return

        self.waiting = [w for w in self.waiting if w[0] != id(owner)]
        self._addOwner(owner, access)

        debuglog(f" {self} is claimed '{access.mode}', {access.count} units")

    def subscribeToReleases(self, callback):
        """Schedule C{callback} to be invoked every time this lock is
        released.  Returns a L{Subscription}."""
        return self.release_subs.subscribe(callback)

    def release(self, owner, access):
        """Release the lock"""
        assert isinstance(access, LockAccess)

        if not access.count:
            return

        debuglog(f"{self} release({owner}, {access.mode}, {access.count})")
        if not self._removeOwner(owner, access):
            debuglog(f"{self} already released")
            return

        self._tryWakeUp()

        # notify any listeners
        self.release_subs.deliver()

    def _tryWakeUp(self):
        # After an exclusive access, we may need to wake up several waiting.
        # Break out of the loop when the first waiting client should not be
        # awakened.
        num_excl, num_counting = self._claimed_excl, self._claimed_counting
        for i, (w_owner_id, w_access, d) in enumerate(self.waiting):
            if w_access.mode == 'counting':
                if num_excl > 0 or num_counting >= self.maxCount:
                    break
                num_counting = num_counting + w_access.count
            else:
                # w_access.mode == 'exclusive'
                if num_excl > 0 or num_counting > 0:
                    break
                num_excl = num_excl + w_access.count

            # If the waiter has a deferred, wake it up and clear the deferred
            # from the wait queue entry to indicate that it has been woken.
            if d:
                self.waiting[i] = (w_owner_id, w_access, None)
                eventually(d.callback, self)

    def waitUntilMaybeAvailable(self, owner, access):
        """Fire when the lock *might* be available. The deferred may be fired spuriously and
        the lock is not necessarily available, thus the caller will need to check with
        isAvailable() when the deferred fires.

        A single requester must not have more than one pending waitUntilMaybeAvailable() on a
        single lock.

        The caller must guarantee, that once the returned deferred is fired, either the lock is
        checked for availability and claimed if it's available, or the it is indicated as no
        longer interesting by calling stopWaitingUntilAvailable(). The caller does not need to
        do this immediately after deferred is fired, an eventual execution is sufficient.
        """
        debuglog(f"{self} waitUntilAvailable({owner})")
        assert isinstance(access, LockAccess)
        if self.isAvailable(owner, access):
            return defer.succeed(self)
        d = defer.Deferred()

        # Are we already in the wait queue?
        w_index = self._find_waiting(owner)
        if w_index is not None:
            _, _, old_d = self.waiting[w_index]
            assert old_d is None, (
                "waitUntilMaybeAvailable() must not be called again before the "
                "previous deferred fired"
            )
            self.waiting[w_index] = (id(owner), access, d)
        else:
            self.waiting.append((id(owner), access, d))
        return d

    def stopWaitingUntilAvailable(self, owner, access, d):
        """Stop waiting for lock to become available. `d` must be the result of a previous call
        to `waitUntilMaybeAvailable()`. If `d` has not been woken up already by calling its
        callback, it will be done as part of this function
        """
        debuglog(f"{self} stopWaitingUntilAvailable({owner})")
        assert isinstance(access, LockAccess)

        w_index = self._find_waiting(owner)
        assert w_index is not None, "The owner was not waiting for the lock"
        _, _, old_d = self.waiting[w_index]
        if old_d is not None:
            assert d is old_d, "The supplied deferred must be a result of waitUntilMaybeAvailable()"
            del self.waiting[w_index]
            d.callback(None)
        else:
            del self.waiting[w_index]
            # if the callback has already been woken up, then it must schedule another waiter,
            # otherwise we will have an available lock with a waiter list and no-one to wake the
            # waiters up.
            self._tryWakeUp()

    def isOwner(self, owner, access):
        return (id(owner), access) in self.owners


class RealMasterLock(BaseLock, service.SharedService):
    def __init__(self, name):
        # the caller will want to call updateFromLockId after initialization
        super().__init__(name, 0)
        self.config_version = -1
        self._updateDescription()

    def _updateDescription(self):
        self.description = f"<MasterLock({self.lockName}, {self.maxCount})>"

    def getLockForWorker(self, workername):
        return self

    def updateFromLockId(self, lockid, config_version):
        assert self.lockName == lockid.name
        assert isinstance(config_version, int)

        self.config_version = config_version
        self.setMaxCount(lockid.maxCount)
        self._updateDescription()


class RealWorkerLock(service.SharedService):
    def __init__(self, name):
        super().__init__()

        # the caller will want to call updateFromLockId after initialization
        self.lockName = name
        self.maxCount = None
        self.maxCountForWorker = None
        self.config_version = -1
        self._updateDescription()
        self.locks = {}

    def __repr__(self):
        return self.description

    def getLockForWorker(self, workername):
        if workername not in self.locks:
            maxCount = self.maxCountForWorker.get(workername, self.maxCount)
            lock = self.locks[workername] = BaseLock(self.lockName, maxCount)
            self._updateDescriptionForLock(lock, workername)
            self.locks[workername] = lock
        return self.locks[workername]

    def _updateDescription(self):
        self.description = (
            f"<WorkerLock({self.lockName}, {self.maxCount}, {self.maxCountForWorker})>"
        )

    def _updateDescriptionForLock(self, lock, workername):
        lock.description = (
            f"<WorkerLock({lock.lockName}, {lock.maxCount})[{workername}] {id(lock)}>"
        )

    def updateFromLockId(self, lockid, config_version):
        assert self.lockName == lockid.name
        assert isinstance(config_version, int)

        self.config_version = config_version

        self.maxCount = lockid.maxCount
        self.maxCountForWorker = lockid.maxCountForWorker

        self._updateDescription()

        for workername, lock in self.locks.items():
            maxCount = self.maxCountForWorker.get(workername, self.maxCount)
            lock.setMaxCount(maxCount)
            self._updateDescriptionForLock(lock, workername)


class LockAccess(util.ComparableMixin):
    """I am an object representing a way to access a lock.

    @param lockid: LockId instance that should be accessed.
    @type  lockid: A MasterLock or WorkerLock instance.

    @param mode: Mode of accessing the lock.
    @type  mode: A string, either 'counting' or 'exclusive'.

    @param count: How many units does the access occupy
    @type  count: Integer, not negative, default is 1 for backwards
                  compatibility
    """

    compare_attrs: ClassVar[Sequence[str]] = ('lockid', 'mode', 'count')

    def __init__(self, lockid, mode, count=1):
        self.lockid = lockid
        self.mode = mode
        self.count = count

        assert isinstance(lockid, (MasterLock, WorkerLock))
        assert mode in ['counting', 'exclusive']
        assert isinstance(count, int)
        if mode == 'exclusive':
            assert count == 1
        else:
            assert count >= 0


class BaseLockId(util.ComparableMixin):
    """Abstract base class for LockId classes.

    Sets up the 'access()' function for the LockId's available to the user
    (MasterLock and WorkerLock classes).
    Derived classes should add
    - Comparison with the L{util.ComparableMixin} via the L{compare_attrs}
      class variable.
    - Link to the actual lock class should be added with the L{lockClass}
      class variable.
    """

    def access(self, mode, count=1):
        """Express how the lock should be accessed"""
        assert mode in ['counting', 'exclusive']
        assert isinstance(count, int)
        assert count >= 0
        return LockAccess(self, mode, count)

    def defaultAccess(self):
        """For buildbot 0.7.7 compatibility: When user doesn't specify an access
        mode, this one is chosen.
        """
        return self.access('counting')


# master.cfg should only reference the following MasterLock and WorkerLock
# classes. They are identifiers that will be turned into real Locks later,
# via the BotMaster.getLockByID method.
class MasterLock(BaseLockId):
    """I am a semaphore that limits the number of simultaneous actions.

    Builds and BuildSteps can declare that they wish to claim me as they run.
    Only a limited number of such builds or steps will be able to run
    simultaneously. By default this number is one, but my maxCount parameter
    can be raised to allow two or three or more operations to happen at the
    same time.

    Use this to protect a resource that is shared among all builders and all
    workers, for example to limit the load on a common SVN repository.
    """

    compare_attrs: ClassVar[Sequence[str]] = ('name', 'maxCount')
    lockClass = RealMasterLock

    def __init__(self, name, maxCount=1):
        self.name = name
        self.maxCount = maxCount


class WorkerLock(BaseLockId):
    """I am a semaphore that limits simultaneous actions on each worker.

    Builds and BuildSteps can declare that they wish to claim me as they run.
    Only a limited number of such builds or steps will be able to run
    simultaneously on any given worker. By default this number is one,
    but my maxCount parameter can be raised to allow two or three or more
    operations to happen on a single worker at the same time.

    Use this to protect a resource that is shared among all the builds taking
    place on each worker, for example to limit CPU or memory load on an
    underpowered machine.

    Each worker will get an independent copy of this semaphore. By
    default each copy will use the same owner count (set with maxCount), but
    you can provide maxCountForWorker with a dictionary that maps workername to
    owner count, to allow some workers more parallelism than others.

    """

    compare_attrs: ClassVar[Sequence[str]] = ('name', 'maxCount', '_maxCountForWorkerList')
    lockClass = RealWorkerLock

    def __init__(self, name, maxCount=1, maxCountForWorker=None):
        self.name = name
        self.maxCount = maxCount
        if maxCountForWorker is None:
            maxCountForWorker = {}
        self.maxCountForWorker = maxCountForWorker
        # for comparison purposes, turn this dictionary into a stably-sorted
        # list of tuples
        self._maxCountForWorkerList = tuple(sorted(self.maxCountForWorker.items()))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This is a static resource type and set of endpoints used as common data by
# tests.
from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

testData = {
    13: {'testid': 13, 'info': 'ok', 'success': True, 'tags': []},
    14: {'testid': 14, 'info': 'failed', 'success': False, 'tags': []},
    15: {
        'testid': 15,
        'info': 'warned',
        'success': True,
        'tags': [
            'a',
            'b',
        ],
    },
    16: {'testid': 16, 'info': 'skipped', 'success': True, 'tags': ['a']},
    17: {'testid': 17, 'info': 'ignored', 'success': True, 'tags': []},
    18: {'testid': 18, 'info': 'unexp', 'success': False, 'tags': []},
    19: {'testid': 19, 'info': 'todo', 'success': True, 'tags': []},
    20: {'testid': 20, 'info': 'error', 'success': False, 'tags': []},
}
stepData = {
    13: {'stepid': 13, 'testid': 13, 'info': 'ok'},
    14: {'stepid': 14, 'testid': 13, 'info': 'failed'},
    15: {'stepid': 15, 'testid': 14, 'info': 'failed'},
}


class TestsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/tests",
        "/test",
    ]
    rootLinkName = 'tests'

    def get(self, resultSpec, kwargs):
        # results are sorted by ID for test stability
        return defer.succeed(sorted(testData.values(), key=lambda v: v['testid']))


class RawTestsEndpoint(base.Endpoint):
    kind = base.EndpointKind.RAW
    pathPatterns = [
        "/rawtest",
    ]

    def get(self, resultSpec, kwargs):
        return defer.succeed({"filename": "test.txt", "mime-type": "text/test", 'raw': 'value'})


class FailEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/test/fail",
    ]

    def get(self, resultSpec, kwargs):
        return defer.fail(RuntimeError('oh noes'))


class TestEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/tests/n:testid",
        "/test/n:testid",
    ]

    def get(self, resultSpec, kwargs):
        if kwargs['testid'] == 0:
            return None
        return defer.succeed(testData[kwargs['testid']])

    def control(self, action, args, kwargs):
        if action == "fail":
            return defer.fail(RuntimeError("oh noes"))
        return defer.succeed({'action': action, 'args': args, 'kwargs': kwargs})


class StepsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/tests/n:testid/steps",
    ]

    def get(self, resultSpec, kwargs):
        data = [step for step in stepData.values() if step['testid'] == kwargs['testid']]
        # results are sorted by ID for test stability
        return defer.succeed(sorted(data, key=lambda v: v['stepid']))


class StepEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/tests/n:testid/steps/n:stepid",
    ]

    def get(self, resultSpec, kwargs):
        if kwargs['testid'] == 0:
            return None
        return defer.succeed(testData[kwargs['testid']])


class Step(base.ResourceType):
    name = "step"
    plural = "steps"
    endpoints = [StepsEndpoint, StepEndpoint]

    class EntityType(types.Entity):
        stepid = types.Integer()
        testid = types.Integer()
        info = types.String()

    entityType = EntityType(name)


class Test(base.ResourceType):
    name = "test"
    plural = "tests"
    endpoints = [TestsEndpoint, TestEndpoint, FailEndpoint, RawTestsEndpoint]

    class EntityType(types.Entity):
        testid = types.Integer()
        info = types.String()
        success = types.Boolean()
        tags = types.List(of=types.String())

    entityType = EntityType(name)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# Utility functions for catching Python warnings.
# Twisted's TestCase already gathers produced warnings
# (see t.t.u.T.flushWarnings()), however Twisted's implementation doesn't
# allow fine-grained control over caught warnings.
# This implementation uses context wrapper style to specify interesting
# block of code to catch warnings, which allows to easily specify which
# exactly statements should generate warnings and which shouldn't.
# Also this implementation allows nested checks.


import contextlib
import re
import warnings


@contextlib.contextmanager
def _recordWarnings(category, output):
    assert isinstance(output, list)

    unrelated_warns = []
    with warnings.catch_warnings(record=True) as all_warns:
        # Cause all warnings of the provided category to always be
        # triggered.
        warnings.simplefilter("always", category)

        yield

        # Filter warnings.
        for w in all_warns:
            if isinstance(w.message, category):
                output.append(w)
            else:
                unrelated_warns.append(w)

    # Re-raise unrelated warnings.
    for w in unrelated_warns:
        warnings.warn_explicit(w.message, w.category, w.filename, w.lineno)


@contextlib.contextmanager
def assertProducesWarnings(
    filter_category, num_warnings=None, messages_patterns=None, message_pattern=None
):
    if messages_patterns is not None:
        assert message_pattern is None
        assert num_warnings is None
        num_warnings = len(messages_patterns)
    else:
        assert num_warnings is not None or message_pattern is not None

    warns = []
    with _recordWarnings(filter_category, warns):
        yield

    if num_warnings is None:
        num_warnings = 1

    warns_str = '\n'.join(map(str, warns))
    assert len(warns) == num_warnings, (
        "Number of occurred warnings is not correct. "
        f"Expected {num_warnings} warnings, received {len(warns)}:\n"
        f"{warns_str}"
    )

    if messages_patterns is None and message_pattern is not None:
        messages_patterns = [message_pattern] * num_warnings

    if messages_patterns is not None:
        for w, pattern in zip(warns, messages_patterns):
            # TODO: Maybe don't use regexp, but use simple substring check?
            warns_str = '\n'.join(map(str, warns))
            assert re.search(pattern, str(w.message)), (
                "Warning pattern doesn't match. Expected pattern:\n"
                f"{pattern}\n"
                "Received message:\n"
                f"{w.message}\n"
                "All gathered warnings:\n"
                f"{warns_str}"
            )


@contextlib.contextmanager
def assertProducesWarning(filter_category, message_pattern=None):
    with assertProducesWarnings(filter_category, num_warnings=1, message_pattern=message_pattern):
        yield


@contextlib.contextmanager
def assertNotProducesWarnings(filter_category):
    with assertProducesWarnings(filter_category, 0):
        yield


@contextlib.contextmanager
def ignoreWarning(category):
    with _recordWarnings(category, []):
        yield



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#
# code inspired/copied from contrib/github_buildbot
#  and inspired from code from the Chromium project
# otherwise, Andrew Melo <andrew.melo@gmail.com> wrote the rest
# but "the rest" is pretty minimal


from __future__ import annotations

import json
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from twisted.internet import defer

from buildbot.util import bytes2unicode

if TYPE_CHECKING:
    from twisted.web.server import Request

    from buildbot.master import BuildMaster


class BaseHookHandler:
    def __init__(self, master: BuildMaster, options: dict[str, Any] | None):
        self.master = master
        self.options = options

    def getChanges(
        self, request: Request
    ) -> defer.Deferred[tuple[list[dict[str, Any]], str | None]]:
        """
        Consumes a naive build notification (the default for now)
        basically, set POST variables to match commit object parameters:
        revision, revlink, comments, branch, who, files, links

        files, links and properties will be de-json'd, the rest are interpreted as strings
        """

        def firstOrNothing(value: Any) -> str:
            """
            Small helper function to return the first value (if value is a list)
            or return the whole thing otherwise.

            Make sure to properly decode bytes to unicode strings.
            """
            if isinstance(value, type([])):
                value = value[0]  # type: ignore[index]
            return bytes2unicode(value)

        args = cast(dict[bytes, list[bytes]], request.args)
        # first, convert files, links and properties
        files = None
        if args.get(b'files'):
            files = json.loads(firstOrNothing(args.get(b'files')))
        else:
            files = []

        properties = None
        if args.get(b'properties'):
            properties = json.loads(firstOrNothing(args.get(b'properties')))
        else:
            properties = {}

        revision = firstOrNothing(args.get(b'revision'))
        when: str | float | None = firstOrNothing(args.get(b'when_timestamp'))
        if when is None:
            when = firstOrNothing(args.get(b'when'))
        if when is not None:
            when = float(when)
        author = firstOrNothing(args.get(b'author'))
        if not author:
            author = firstOrNothing(args.get(b'who'))
        committer = firstOrNothing(args.get(b'committer'))
        comments = firstOrNothing(args.get(b'comments'))
        branch = firstOrNothing(args.get(b'branch'))
        category = firstOrNothing(args.get(b'category'))
        revlink = firstOrNothing(args.get(b'revlink'))
        repository = firstOrNothing(args.get(b'repository')) or ''
        project = firstOrNothing(args.get(b'project')) or ''
        codebase = firstOrNothing(args.get(b'codebase'))

        chdict = {
            "author": author,
            "committer": committer,
            "files": files,
            "comments": comments,
            "revision": revision,
            "when_timestamp": when,
            "branch": branch,
            "category": category,
            "revlink": revlink,
            "properties": properties,
            "repository": repository,
            "project": project,
            "codebase": codebase,
        }
        return defer.succeed(([chdict], None))


base = BaseHookHandler  # alternate name for buildbot plugin



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import re
import traceback
import warnings

import yaml
from evalidate import Expr
from evalidate import base_eval_model
from twisted.internet import defer

from buildbot.db.model import Model
from buildbot.plugins import util
from buildbot.plugins.db import get_plugins
from buildbot.process import buildstep
from buildbot.process.properties import Properties
from buildbot.process.properties import renderer
from buildbot.process.results import SUCCESS
from buildbot.steps.shell import ShellCommand
from buildbot.steps.trigger import Trigger
from buildbot.steps.worker import CompositeStepMixin


class BuildbotCiYmlInvalid(Exception):
    pass


_env_string_key_re = re.compile(r'^\s*(\w+)=')
_env_string_value_re = re.compile(r'''(?:"((?:\\.|[^"])*?)"|'([^']*?)'|(\S*))''')


def parse_env_string(env_str, parent_env=None):
    env_str = env_str.strip()
    orig_env_str = env_str

    props = {}
    if parent_env:
        props.update(parent_env)
    if not env_str:
        return props

    while env_str:
        m = _env_string_key_re.match(env_str)
        if m is None:
            raise ValueError(f'Could not parse \'{orig_env_str}\': splitting \'{env_str}\' failed')
        k = m.group(1)

        env_str = env_str[m.end() :]

        m = _env_string_value_re.match(env_str)
        if m is None:
            raise ValueError(f'Could not parse \'{orig_env_str}\': splitting \'{env_str}\' failed')

        env_str = env_str[m.end() :]

        v = m.group(1) or m.group(2) or m.group(3) or ''
        props[k] = v

    return props


def interpolate_constructor(loader, node):
    value = loader.construct_scalar(node)
    return util.Interpolate(value)


class BuildbotCiLoader(yaml.SafeLoader):
    constructors_loaded = False

    @classmethod
    def ensure_constructors_loaded(cls):
        if cls.constructors_loaded:
            return
        cls.load_constructors()

    @classmethod
    def load_constructors(cls):
        cls.add_constructor('!Interpolate', interpolate_constructor)
        cls.add_constructor('!i', interpolate_constructor)

        steps = get_plugins('steps', None, load_now=True)
        for step_name in steps.names:
            # Accessing a step from the plugin DB may raise warrings (e.g. deprecation).
            # We don't want them logged until the step is actually used.
            with warnings.catch_warnings(record=True) as all_warnings:
                warnings.simplefilter("always")
                step_class = steps.get(step_name)
                step_warnings = list(all_warnings)
            cls.register_step_class(step_name, step_class, step_warnings)

    @classmethod
    def register_step_class(cls, name, step_class, step_warnings):
        def step_constructor(loader, node):
            try:
                if isinstance(node, yaml.ScalarNode):
                    args = [loader.construct_scalar(node)]
                    kwargs = {}
                elif isinstance(node, yaml.SequenceNode):
                    args = loader.construct_sequence(node)
                    kwargs = {}
                elif isinstance(node, yaml.MappingNode):
                    args = []
                    kwargs = loader.construct_mapping(node)
                else:
                    raise Exception('Unsupported node type')
            except Exception as e:
                raise Exception(f"Could not parse steps arguments: {e}") from e

            # Re-raise all warnings that occurred when accessing step class. We only want them to be
            # logged if the configuration actually uses the step class.
            for w in step_warnings:
                warnings.warn_explicit(w.message, w.category, w.filename, w.lineno)

            return step_class(*args, **kwargs)

        cls.add_constructor('!' + name, step_constructor)


class BuildbotCiYml:
    SCRIPTS = (
        "before_install",
        "install",
        "after_install",
        "before_script",
        "script",
        "after_script",
    )

    steps_loaded = False

    def __init__(self):
        self.config_dict = None
        self.label_mapping = {}
        self.global_env = {}
        self.script_commands = {}
        for script in self.SCRIPTS:
            self.script_commands[script] = []
        self.matrix = []

    @classmethod
    def load_from_str(cls, config_input):
        BuildbotCiLoader.ensure_constructors_loaded()

        try:
            config_dict = yaml.load(config_input, Loader=BuildbotCiLoader)
        except Exception as e:
            raise BuildbotCiYmlInvalid(f"Invalid YAML data\n{e}") from e

        return cls.load_from_dict(config_dict)

    @classmethod
    def load_from_dict(cls, config):
        yml = cls()
        yml.load_from_dict_internal(config)
        return yml

    def load_from_dict_internal(self, config):
        self.config = config
        self.label_mapping = self.config.get('label_mapping', {})
        self.global_env = BuildbotCiYml.load_global_env(config)
        self.script_commands = BuildbotCiYml.load_scripts(config)
        self.matrix = BuildbotCiYml.load_matrix(config, self.global_env)

    @classmethod
    def load_global_env(cls, config):
        env = config.get("env", None)

        if env is None:
            return {}

        if isinstance(env, list):
            return {}

        if isinstance(env, dict):
            env = env.get('global')
            if isinstance(env, str):
                return parse_env_string(env)
            if isinstance(env, list):
                global_env = {}
                for e in env:
                    global_env.update(parse_env_string(e))
                return global_env
            raise BuildbotCiYmlInvalid("'env.global' configuration parameter is invalid")

        raise BuildbotCiYmlInvalid("'env' parameter is invalid")

    @classmethod
    def load_scripts(cls, config):
        script_commands = {}
        for script in cls.SCRIPTS:
            commands = config.get(script, [])
            if isinstance(commands, str):
                commands = [commands]
            if not isinstance(commands, list):
                raise BuildbotCiYmlInvalid(f"'{script}' parameter is invalid")
            script_commands[script] = commands
        return script_commands

    @staticmethod
    def adjust_matrix_config(c, global_env):
        c = c.copy()
        c['env'] = parse_env_string(c.get('env', ''), global_env)
        return c

    @classmethod
    def load_matrix(cls, config, global_env):
        return [
            cls.adjust_matrix_config(matrix_config, global_env)
            for matrix_config in config.get('matrix', {}).get('include') or []
        ]


class BuildbotTestCiReadConfigMixin:
    config_filenames = ['.bbtravis.yml', '.buildbot-ci.yml']
    config = None

    @defer.inlineCallbacks
    def get_config_yml_from_worker(self):
        exceptions = []
        for filename in self.config_filenames:
            try:
                config_yml = yield self.getFileContentFromWorker(filename, abandonOnFailure=True)
                return filename, config_yml
            except buildstep.BuildStepFailed as e:
                exceptions.append(e)

        return None, exceptions

    @defer.inlineCallbacks
    def get_ci_config(self):
        filename, result = yield self.get_config_yml_from_worker()
        if not filename:
            exceptions = result
            msg = ' '.join(str(exceptions))

            self.descriptionDone = "failed to read configuration"
            self.addCompleteLog(
                'error',
                f'Failed to read configuration from files {self.config_filenames}: got {msg}',
            )
            raise buildstep.BuildStepFailed("failed to read configuration")

        config_yml = result

        self.addCompleteLog(filename, config_yml)

        try:
            config = BuildbotCiYml.load_from_str(config_yml)
        except BuildbotCiYmlInvalid as e:
            self.descriptionDone = f'bad configuration file {filename}'
            self.addCompleteLog('error', f'Bad configuration file:\n{e}')
            raise buildstep.BuildStepFailed(f'bad configuration file {filename}') from e

        return config


class BuildbotTestCiTrigger(BuildbotTestCiReadConfigMixin, CompositeStepMixin, Trigger):
    def __init__(self, scheduler, **kwargs):
        super().__init__(
            name='buildbot-test-ci trigger',
            waitForFinish=True,
            schedulerNames=[scheduler],
            haltOnFailure=True,
            flunkOnFailure=True,
            sourceStamps=[],
            alwaysUseLatest=False,
            updateSourceStamp=False,
            **kwargs,
        )

    @defer.inlineCallbacks
    def run(self):
        self.config = yield self.get_ci_config()

        rv = yield super().run()
        return rv

    def _replace_label(self, v):
        return str(self.config.label_mapping.get(v, v))

    def build_scheduler_for_env(self, scheduler_name, env):
        new_build_props = Properties()
        new_build_props.setProperty(
            "BUILDBOT_PULL_REQUEST", self.getProperty("BUILDBOT_PULL_REQUEST"), "inherit"
        )
        for k, v in env.items():
            if k == "env":
                # v is dictionary
                new_build_props.update(v, "BuildbotTestCiTrigger")
            else:
                new_build_props.setProperty(k, v, "BuildbotTestCiTrigger")

        tags_from_props = sorted(
            f'{self._replace_label(k)}:{self._replace_label(v)}'
            for k, (v, _) in new_build_props.asDict().items()
            if k not in self.config.global_env.keys() and k != 'BUILDBOT_PULL_REQUEST'
        )

        tags = [t for t in self.build.builder.config.tags if t not in ('try', 'spawner')]

        new_build_props.setProperty(
            "virtual_builder_name", " ".join(tags + tags_from_props), "BuildbotTestCiTrigger"
        )
        new_build_props.setProperty(
            "virtual_builder_tags", tags + tags_from_props, "BuildbotTestCiTrigger"
        )
        new_build_props.setProperty(
            "matrix_label", "/".join(tags_from_props), "BuildbotTestCiTrigger"
        )

        return (scheduler_name, new_build_props)

    def createTriggerProperties(self, props):
        return props

    def getSchedulersAndProperties(self):
        scheduler_name = self.schedulerNames[0]
        return [self.build_scheduler_for_env(scheduler_name, env) for env in self.config.matrix]


eval_model = base_eval_model.clone()
eval_model.nodes.append('Mul')
eval_model.nodes.append('Slice')
eval_model.nodes.append('Tuple')


def evaluate_condition(condition, local_dict):
    expr = Expr(condition, eval_model)
    return bool(expr.eval(local_dict))


class BuildbotCiSetupSteps(BuildbotTestCiReadConfigMixin, CompositeStepMixin, buildstep.BuildStep):
    name = "setup-steps"
    haltOnFailure = True
    flunkOnFailure = True
    MAX_NAME_LENGTH = 47
    disable = False

    def _add_step(self, command):
        name = None
        condition = None
        step = None
        original_command = command
        if isinstance(command, dict):
            name = command.get("title")
            condition = command.get("condition")
            step = command.get("step")
            command = command.get("cmd")

        if isinstance(command, buildstep.BuildStep):
            step = command

        if condition is not None:
            try:
                local_dict = {k: v for k, (v, s) in self.build.getProperties().properties.items()}
                if not evaluate_condition(condition, local_dict):
                    return
            except Exception:
                self.descriptionDone = "Problem parsing condition"
                self.addCompleteLog("condition error", traceback.format_exc())
                return

        if step is None:
            if command is None:
                self.addCompleteLog(
                    "BuildbotCiSetupSteps error",
                    f"Neither step nor cmd is defined: {original_command}",
                )
                return
            if name is None:
                name = self._name_from_command(command)

            @renderer
            def render_env(props):
                return {str(k): str(v[0]) for k, v in props.properties.items()}

            step = ShellCommand(
                name=name,
                description=command,
                command=command,
                doStepIf=not self.disable,
                env=render_env,
            )
        self.build.addStepsAfterLastStep([step])

    def _name_from_command(self, name):
        name = name.lstrip("#").lstrip(" ").split("\n")[0]
        max_length = Model.steps.c.name.type.length
        if len(name) > max_length:
            name = name[: max_length - 3] + "..."
        return name

    @defer.inlineCallbacks
    def run(self):
        config = yield self.get_ci_config()
        for k in BuildbotCiYml.SCRIPTS:
            for command in config.script_commands[k]:
                self._add_step(command=command)

        return SUCCESS



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import contextlib
from typing import Any

from twisted.internet import defer
from twisted.python import log
from twisted.spread import pb

from buildbot.pbutil import decode
from buildbot.util import deferwaiter
from buildbot.worker.protocols import base


class Listener(base.UpdateRegistrationListener):
    name = "pbListener"

    def __init__(self, master):
        super().__init__()
        self.ConnectionClass = Connection
        self.master = master

    def get_manager(self):
        return self.master.pbmanager

    def before_connection_setup(self, mind, workerName):
        log.msg(f"worker '{workerName}' attaching from {mind.broker.transport.getPeer()}")
        try:
            mind.broker.transport.setTcpKeepAlive(1)
        except Exception:
            log.err("Can't set TcpKeepAlive")


class ReferenceableProxy(pb.Referenceable):
    def __init__(self, impl):
        assert isinstance(impl, self.ImplClass)
        self.impl = impl

    def __getattr__(self, name):
        return getattr(self.impl, name)


# Proxy are just ReferenceableProxy to the Impl classes
class RemoteCommand(ReferenceableProxy):
    ImplClass = base.RemoteCommandImpl


class FileReaderProxy(ReferenceableProxy):
    ImplClass = base.FileReaderImpl


class FileWriterProxy(ReferenceableProxy):
    ImplClass = base.FileWriterImpl


class _NoSuchMethod(Exception):
    """Rewrapped pb.NoSuchMethod remote exception"""


@contextlib.contextmanager
def _wrapRemoteException():
    try:
        yield
    except pb.RemoteError as e:
        if e.remoteType in (
            b'twisted.spread.flavors.NoSuchMethod',
            'twisted.spread.flavors.NoSuchMethod',
        ):
            raise _NoSuchMethod(e) from e
        raise


class Connection(base.Connection, pb.Avatar):
    proxies = {base.FileWriterImpl: FileWriterProxy, base.FileReaderImpl: FileReaderProxy}
    # TODO: configure keepalive_interval in
    # c['protocols']['pb']['keepalive_interval']
    keepalive_timer: None = None
    keepalive_interval = 3600
    info: Any = None

    def __init__(self, master, worker, mind):
        super().__init__(worker.workername)
        self.master = master
        self.worker = worker
        self.mind = mind
        self._keepalive_waiter = deferwaiter.DeferWaiter()
        self._keepalive_action_handler = deferwaiter.RepeatedActionHandler(
            master.reactor, self._keepalive_waiter, self.keepalive_interval, self._do_keepalive
        )

    # methods called by the PBManager

    @defer.inlineCallbacks
    def attached(self, mind):
        self.startKeepaliveTimer()
        self.notifyOnDisconnect(self._stop_keepalive_timer)
        # pbmanager calls perspective.attached; pass this along to the
        # worker
        yield self.worker.attached(self)
        # and then return a reference to the avatar
        return self

    def detached(self, mind):
        self.stopKeepaliveTimer()
        self.mind = None
        self.notifyDisconnected()

    # disconnection handling
    @defer.inlineCallbacks
    def _stop_keepalive_timer(self):
        self.stopKeepaliveTimer()
        yield self._keepalive_waiter.wait()

    def loseConnection(self):
        self.stopKeepaliveTimer()
        tport = self.mind.broker.transport
        # this is the polite way to request that a socket be closed
        tport.loseConnection()
        try:
            # but really we don't want to wait for the transmit queue to
            # drain. The remote end is unlikely to ACK the data, so we'd
            # probably have to wait for a (20-minute) TCP timeout.
            # tport._closeSocket()
            # however, doing _closeSocket (whether before or after
            # loseConnection) somehow prevents the notifyOnDisconnect
            # handlers from being run. Bummer.
            tport.offset = 0
            tport.dataBuffer = b""
        except Exception:
            # however, these hacks are pretty internal, so don't blow up if
            # they fail or are unavailable
            log.msg("failed to accelerate the shutdown process")

    # keepalive handling

    def _do_keepalive(self):
        return self.mind.callRemote('print', message="keepalive")

    def stopKeepaliveTimer(self):
        self._keepalive_action_handler.stop()

    def startKeepaliveTimer(self):
        assert self.keepalive_interval
        self._keepalive_action_handler.start()

    # methods to send messages to the worker

    def remotePrint(self, message):
        return self.mind.callRemote('print', message=message)

    @defer.inlineCallbacks
    def remoteGetWorkerInfo(self):
        try:
            with _wrapRemoteException():
                # Try to call buildbot-worker method.
                info = yield self.mind.callRemote('getWorkerInfo')
            return decode(info)
        except _NoSuchMethod:
            yield self.remotePrint(
                "buildbot-slave detected, failing back to deprecated buildslave API. "
                "(Ignoring missing getWorkerInfo method.)"
            )
            info = {}

            # Probably this is deprecated buildslave.
            log.msg(
                "Worker.getWorkerInfo is unavailable - falling back to deprecated buildslave API"
            )

            try:
                with _wrapRemoteException():
                    info = yield self.mind.callRemote('getSlaveInfo')
            except _NoSuchMethod:
                log.msg("Worker.getSlaveInfo is unavailable - ignoring")

            # newer workers send all info in one command
            if "slave_commands" in info:
                assert "worker_commands" not in info
                info["worker_commands"] = info.pop("slave_commands")
                return info

            # Old version buildslave - need to retrieve list of supported
            # commands and version using separate requests.
            try:
                with _wrapRemoteException():
                    info["worker_commands"] = yield self.mind.callRemote('getCommands')
            except _NoSuchMethod:
                log.msg("Worker.getCommands is unavailable - ignoring")

            try:
                with _wrapRemoteException():
                    info["version"] = yield self.mind.callRemote('getVersion')
            except _NoSuchMethod:
                log.msg("Worker.getVersion is unavailable - ignoring")

            return decode(info)

    @defer.inlineCallbacks
    def remoteSetBuilderList(self, builders):
        builders = yield self.mind.callRemote('setBuilderList', builders)
        self.builders = builders
        return builders

    def remoteStartCommand(self, remoteCommand, builderName, commandId, commandName, args):
        workerforbuilder = self.builders.get(builderName)
        remoteCommand = RemoteCommand(remoteCommand)
        args = self.createArgsProxies(args)
        return workerforbuilder.callRemote(
            'startCommand', remoteCommand, commandId, commandName, args
        )

    @defer.inlineCallbacks
    def remoteShutdown(self):
        # First, try the "new" way - calling our own remote's shutdown
        # method. The method was only added in 0.8.3, so ignore NoSuchMethod
        # failures.
        @defer.inlineCallbacks
        def new_way():
            try:
                with _wrapRemoteException():
                    yield self.mind.callRemote('shutdown')
                    # successful shutdown request
                    return True
            except _NoSuchMethod:
                # fall through to the old way
                return False

            except pb.PBConnectionLost:
                # the worker is gone, so call it finished
                return True

        if (yield new_way()):
            return  # done!

        # Now, the old way. Look for a builder with a remote reference to the
        # client side worker. If we can find one, then call "shutdown" on the
        # remote builder, which will cause the worker buildbot process to exit.
        def old_way():
            d = None
            for b in self.worker.workerforbuilders.values():
                if b.remote:
                    d = b.mind.callRemote("shutdown")
                    break

            if d:
                name = self.worker.workername
                log.msg(f"Shutting down (old) worker: {name}")
                # The remote shutdown call will not complete successfully since
                # the buildbot process exits almost immediately after getting
                # the shutdown request.
                # Here we look at the reason why the remote call failed, and if
                # it's because the connection was lost, that means the worker
                # shutdown as expected.

                @d.addErrback
                def _errback(why):
                    if why.check(pb.PBConnectionLost):
                        log.msg(f"Lost connection to {name}")
                    else:
                        log.err(f"Unexpected error when trying to shutdown {name}")

                return d
            log.err("Couldn't find remote builder to shut down worker")
            return defer.succeed(None)

        yield old_way()

    def remoteStartBuild(self, builderName):
        workerforbuilder = self.builders.get(builderName)
        return workerforbuilder.callRemote('startBuild')

    def remoteInterruptCommand(self, builderName, commandId, why):
        workerforbuilder = self.builders.get(builderName)
        return defer.maybeDeferred(workerforbuilder.callRemote, "interruptCommand", commandId, why)

    # perspective methods called by the worker

    def perspective_keepalive(self):
        self.worker.messageReceivedFromWorker()

    def perspective_shutdown(self):
        self.worker.messageReceivedFromWorker()
        self.worker.shutdownRequested()

    def get_peer(self):
        p = self.mind.broker.transport.getPeer()
        return f"{p.host}:{p.port}"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import copy
import json
import os
from collections import OrderedDict

import yaml


# minimalistic raml loader. Support !include tags, and mapping as OrderedDict
class RamlLoader(yaml.SafeLoader):
    pass


def construct_include(loader, node):
    path = os.path.join(os.path.dirname(loader.stream.name), node.value)
    with open(path, encoding='utf-8') as f:
        return yaml.load(f, Loader=RamlLoader)


def construct_mapping(loader, node):
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


RamlLoader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)
RamlLoader.add_constructor('!include', construct_include)


class RamlSpec:
    """
    This class loads the raml specification, and expose useful
    aspects of the spec

    Main usage for now is for the doc, but it can be extended to make sure
    raml spec matches other spec implemented in the tests
    """

    def __init__(self):
        fn = os.path.join(os.path.dirname(__file__), os.pardir, 'spec', 'api.raml')
        with open(fn, encoding='utf-8') as f:
            self.api = yaml.load(f, Loader=RamlLoader)

        with open(fn, encoding='utf-8') as f:
            self.rawraml = f.read()

        endpoints = {}
        self.endpoints_by_type = {}
        self.rawendpoints = {}
        self.endpoints = self.parse_endpoints(endpoints, "", self.api)
        self.types = self.parse_types()

    def parse_endpoints(self, endpoints, base, api, uriParameters=None):
        if uriParameters is None:
            uriParameters = OrderedDict()

        for k, v in api.items():
            if k.startswith("/"):
                ep = base + k
                p = copy.deepcopy(uriParameters)
                if v is not None:
                    p.update(v.get("uriParameters", {}))
                    v["uriParameters"] = p
                    endpoints[ep] = v
                self.parse_endpoints(endpoints, ep, v, p)
            elif k in ['get', 'post']:
                if 'is' not in v:
                    continue

                for _is in v['is']:
                    if not isinstance(_is, dict):
                        raise RuntimeError(f'Unexpected "is" target {type(_is)}: {_is}')

                    if 'bbget' in _is:
                        try:
                            v['eptype'] = _is['bbget']['bbtype']
                        except TypeError as e:
                            raise RuntimeError(f"Unexpected 'is' target {_is['bbget']}") from e

                        self.endpoints_by_type.setdefault(v['eptype'], {})
                        self.endpoints_by_type[v['eptype']][base] = api

                    if 'bbgetraw' in _is:
                        self.rawendpoints.setdefault(base, {})
                        self.rawendpoints[base] = api
        return endpoints

    def reindent(self, s, indent):
        return s.replace("\n", "\n" + " " * indent)

    def format_json(self, j, indent):
        j = json.dumps(j, indent=4).replace(", \n", ",\n")
        return self.reindent(j, indent)

    def parse_types(self):
        types = self.api['types']
        return types

    def iter_actions(self, endpoint):
        ACTIONS_MAGIC = '/actions/'
        for k, v in endpoint.items():
            if k.startswith(ACTIONS_MAGIC):
                k = k[len(ACTIONS_MAGIC) :]
                v = v['post']
                # simplify the raml tree for easier processing
                v['body'] = v['body']['application/json'].get('properties', {})
                yield (k, v)



#!/usr/bin/python


import re
import sys

spaces = re.compile("^ +")


def indent_file(fn: str) -> None:
    lines = []
    with open(fn) as f:
        for line in f:
            lines.append(line)

    def getIndent(i):
        res = spaces.match(lines[i])
        if res is None:
            return 0
        return len(res.group(0))

    def IndentBlock(i, numspaces):
        initIndent = getIndent(i)
        while i < len(lines) and initIndent <= getIndent(i):
            lines[i] = " " * numspaces + lines[i]
            i += 1

    for i, line in enumerate(lines):
        missingIndent = 4 - (getIndent(i) % 4)
        if missingIndent != 4:
            IndentBlock(i, missingIndent)

    with open(fn, 'w') as f:
        for line in lines:
            f.write(line)


def main() -> None:
    for fn in sys.argv[1:]:
        indent_file(fn)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class TestResultSet(Row):
    table = 'test_result_sets'

    id_column = 'id'

    def __init__(
        self,
        id=None,
        builderid=None,
        buildid=None,
        stepid=None,
        description=None,
        category='cat',
        value_unit='unit',
        tests_passed=None,
        tests_failed=None,
        complete=1,
    ):
        super().__init__(
            id=id,
            builderid=builderid,
            buildid=buildid,
            stepid=stepid,
            description=description,
            category=category,
            value_unit=value_unit,
            tests_passed=tests_passed,
            tests_failed=tests_failed,
            complete=complete,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot import util
from buildbot.util import lineboundaries
from buildbot.util.twisted import async_to_deferred


class FakeLogFile:
    def __init__(self, name):
        self.name = name
        self.header = ''
        self.stdout = ''
        self.stderr = ''
        self.lbfs = {}
        self.finished = False
        self._had_errors = False
        self.subPoint = util.subscription.SubscriptionPoint(f"{name!r} log")

    def getName(self):
        return self.name

    def subscribe(self, callback):
        return self.subPoint.subscribe(callback)

    def _getLbf(self, stream):
        try:
            return self.lbfs[stream]
        except KeyError:
            lbf = self.lbfs[stream] = lineboundaries.LineBoundaryFinder()
            return lbf

    def _on_whole_lines(self, stream, lines):
        self.subPoint.deliver(stream, lines)
        assert not self.finished

    def _split_lines(self, stream, text):
        lbf = self._getLbf(stream)
        lines = lbf.append(text)
        if lines is None:
            return
        self._on_whole_lines(stream, lines)

    def addHeader(self, text):
        if not isinstance(text, str):
            text = text.decode('utf-8')
        self.header += text
        self._split_lines('h', text)
        return defer.succeed(None)

    def addStdout(self, text):
        if not isinstance(text, str):
            text = text.decode('utf-8')
        self.stdout += text
        self._split_lines('o', text)
        return defer.succeed(None)

    def addStderr(self, text):
        if not isinstance(text, str):
            text = text.decode('utf-8')
        self.stderr += text
        self._split_lines('e', text)
        return defer.succeed(None)

    def add_header_lines(self, text):
        if not isinstance(text, str):
            text = text.decode('utf-8')
        self.header += text
        self._on_whole_lines('h', text)
        return defer.succeed(None)

    def add_stdout_lines(self, text):
        if not isinstance(text, str):
            text = text.decode('utf-8')
        self.stdout += text
        self._on_whole_lines('o', text)
        return defer.succeed(None)

    def add_stderr_lines(self, text):
        if not isinstance(text, str):
            text = text.decode('utf-8')
        self.stderr += text
        self._on_whole_lines('e', text)
        return defer.succeed(None)

    def had_errors(self):
        return self._had_errors

    def flush(self):
        for stream, lbf in self.lbfs.items():
            lines = lbf.flush()
            if lines is not None:
                self.subPoint.deliver(stream, lines)
        return defer.succeed(None)

    @async_to_deferred
    async def finish(self) -> None:
        assert not self.finished

        await self.flush()
        self.finished = True

        # notify subscribers *after* finishing the log
        self.subPoint.deliver(None, None)

        await self.subPoint.waitForDeliveriesToFinish()
        self._had_errors = len(self.subPoint.pop_exceptions()) > 0

    def fakeData(self, header='', stdout='', stderr=''):
        if header:
            self.header += header
        if stdout:
            self.stdout += stdout
        if stderr:
            self.stderr += stderr



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from buildbot.test.fakedb.row import Row


class Worker(Row):
    table = "workers"

    id_column = 'id'

    def __init__(
        self, id=None, name='some:worker', info=None, paused=0, pause_reason=None, graceful=0
    ):
        if info is None:
            info = {"a": "b"}
        super().__init__(
            id=id, name=name, info=info, paused=paused, pause_reason=pause_reason, graceful=graceful
        )


class ConnectedWorker(Row):
    table = "connected_workers"

    id_column = 'id'

    def __init__(self, id=None, masterid=None, workerid=None):
        super().__init__(id=id, masterid=masterid, workerid=workerid)


class ConfiguredWorker(Row):
    table = "configured_workers"

    id_column = 'id'

    def __init__(self, id=None, buildermasterid=None, workerid=None):
        super().__init__(id=id, buildermasterid=buildermasterid, workerid=workerid)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import re

from twisted.logger import Logger

log = Logger()


class LineBoundaryFinder:
    __slots__ = ['partialLine', 'warned']
    # split at reasonable line length.
    # too big lines will fill master's memory, and slow down the UI too much.
    MAX_LINELENGTH = 4096
    # the lookahead here (`(?=.)`) ensures that `\r` doesn't match at the end
    # of the buffer
    # we also convert cursor control sequence to newlines
    # and ugly \b+ (use of backspace to implement progress bar)
    newline_re = re.compile(r'(\r\n|\r(?=.)|\033\[u|\033\[[0-9]+;[0-9]+[Hf]|\033\[2J|\x08+)')

    def __init__(self, callback=None) -> None:
        self.partialLine: str | None = None
        self.warned: bool = False

    def adjust_line(self, text: str) -> str | None:
        if self.partialLine:
            if len(self.partialLine) > self.MAX_LINELENGTH:
                if not self.warned:
                    # Unfortunately we cannot give more hint as per which log that is
                    log.warn(
                        "Splitting long line: {line_start} {length} "
                        "(not warning anymore for this log)",
                        line_start=self.partialLine[:30],
                        length=len(self.partialLine),
                    )
                    self.warned = True
                # switch the variables, and return previous _partialLine_,
                # split every MAX_LINELENGTH plus a trailing \n
                self.partialLine, text = text, self.partialLine
                ret = []
                while len(text) > self.MAX_LINELENGTH:
                    ret.append(text[: self.MAX_LINELENGTH])
                    text = text[self.MAX_LINELENGTH :]
                ret.append(text)
                result = "\n".join(ret) + "\n"
                return result
            text = self.partialLine + text
            self.partialLine = None
        text = self.newline_re.sub('\n', text)
        if text:
            if text[-1] != '\n':
                i = text.rfind('\n')
                if i >= 0:
                    i = i + 1
                    self.partialLine = text[i:]
                    text = text[:i]
                else:
                    self.partialLine = text
                    return None
            return text
        return None

    def append(self, text: str) -> str | None:
        return self.adjust_line(text)

    def flush(self) -> str | None:
        if self.partialLine is not None:
            return self.append('\n')
        return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db.connector import DBConnector
from buildbot.test.util.db import thd_clean_database
from buildbot.util.sautils import hash_columns
from buildbot.util.twisted import async_to_deferred

from .build_data import BuildData
from .builders import Builder
from .builders import BuilderMaster
from .builders import BuildersTags
from .buildrequests import BuildRequest
from .buildrequests import BuildRequestClaim
from .builds import Build
from .builds import BuildProperty
from .buildsets import Buildset
from .buildsets import BuildsetProperty
from .buildsets import BuildsetSourceStamp
from .changes import Change
from .changes import ChangeFile
from .changes import ChangeProperty
from .changes import ChangeUser
from .changesources import ChangeSource
from .changesources import ChangeSourceMaster
from .codebases import Codebase
from .codebases import CodebaseBranch
from .codebases import CodebaseCommit
from .logs import Log
from .logs import LogChunk
from .masters import Master
from .projects import Project
from .schedulers import Scheduler
from .schedulers import SchedulerChange
from .schedulers import SchedulerMaster
from .sourcestamps import Patch
from .sourcestamps import SourceStamp
from .state import Object
from .state import ObjectState
from .steps import Step
from .tags import Tag
from .test_result_sets import TestResultSet
from .test_results import TestCodePath
from .test_results import TestName
from .test_results import TestResult
from .users import User
from .users import UserInfo
from .workers import ConfiguredWorker
from .workers import ConnectedWorker
from .workers import Worker


def row_bool_to_int(value):
    return 1 if value else 0


class FakeDBConnector(DBConnector):
    """
    A stand-in for C{master.db} that operates without an actual database
    backend.

    The child classes implement various useful assertions and faking methods;
    see their documentation for more.
    """

    MASTER_ID = 824

    def __init__(self, basedir, testcase, auto_upgrade=False, check_version=True, auto_clean=True):
        super().__init__(basedir)
        self.testcase = testcase
        self.auto_upgrade = auto_upgrade
        self.check_version = check_version
        self.auto_clean = auto_clean

    @defer.inlineCallbacks
    def setup(self):
        if self.auto_upgrade:
            yield super().setup(check_version=False)
            yield self.pool.do(thd_clean_database)
            yield self.model.upgrade()
        else:
            yield super().setup(check_version=self.check_version)
            if self.auto_clean:
                yield self.pool.do(thd_clean_database)

    @async_to_deferred
    async def _shutdown(self) -> None:
        await super()._shutdown()
        await self.pool.stop()

    def _match_rows(self, rows, type):
        matched_rows = [r for r in rows if isinstance(r, type)]
        non_matched_rows = [r for r in rows if r not in matched_rows]
        return matched_rows, non_matched_rows

    def _thd_post_insert(self, conn, table):
        if self.pool.engine.dialect.name == 'postgresql':
            # Explicitly inserting primary row IDs does not bump the primary key
            # sequence on Postgres
            autoincrement_foreign_key_column = None

            for column in table.columns.values():
                if (
                    not column.foreign_keys
                    and column.primary_key
                    and isinstance(column.type, sa.Integer)
                ):
                    autoincrement_foreign_key_column = column

            if autoincrement_foreign_key_column is not None:
                r = conn.execute(sa.select(sa.func.max(autoincrement_foreign_key_column)))
                max_column_id = r.fetchone()[0]
                r.close()

                seq_name = f"{table.name}_{autoincrement_foreign_key_column.name}_seq"
                conn.execute(sa.text(f"ALTER SEQUENCE {seq_name} RESTART WITH {max_column_id + 1}"))

    def _thd_maybe_insert_build_data(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildData)
        for row in matched_rows:
            conn.execute(
                self.model.build_data.insert(),
                [
                    {
                        'id': row.id,
                        'buildid': row.buildid,
                        'name': row.name,
                        'value': row.value,
                        'length': row.length,
                        'source': row.source,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.build_data)
        return non_matched_rows

    def _thd_maybe_insert_builder(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Builder)
        for row in matched_rows:
            conn.execute(
                self.model.builders.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                        'projectid': row.projectid,
                        'description': row.description,
                        'description_format': row.description_format,
                        'description_html': row.description_html,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.builders)
        return non_matched_rows

    def _thd_maybe_insert_builder_master(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuilderMaster)
        for row in matched_rows:
            conn.execute(
                self.model.builder_masters.insert(),
                [{'id': row.id, 'builderid': row.builderid, 'masterid': row.masterid}],
            )
        return non_matched_rows

    def _thd_maybe_insert_builder_tags(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildersTags)
        for row in matched_rows:
            conn.execute(
                self.model.builders_tags.insert(),
                [{'builderid': row.builderid, 'tagid': row.tagid}],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.builders_tags)
        return non_matched_rows

    def _thd_maybe_insert_buildrequest(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildRequest)
        for row in matched_rows:
            conn.execute(
                self.model.buildrequests.insert(),
                [
                    {
                        'id': row.id,
                        'buildsetid': row.buildsetid,
                        'builderid': row.builderid,
                        'priority': row.priority,
                        'complete': row_bool_to_int(row.complete),
                        'results': row.results,
                        'submitted_at': row.submitted_at,
                        'complete_at': row.complete_at,
                        'waited_for': row.waited_for,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.buildrequests)
        return non_matched_rows

    def _thd_maybe_insert_buildrequest_claim(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildRequestClaim)
        for row in matched_rows:
            conn.execute(
                self.model.buildrequest_claims.insert(),
                [
                    {
                        'brid': row.brid,
                        'masterid': row.masterid,
                        'claimed_at': row.claimed_at,
                    }
                ],
            )
        return non_matched_rows

    def _thd_maybe_insert_build(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Build)
        for row in matched_rows:
            conn.execute(
                self.model.builds.insert(),
                [
                    {
                        'id': row.id,
                        'number': row.number,
                        'builderid': row.builderid,
                        'buildrequestid': row.buildrequestid,
                        'workerid': row.workerid,
                        'masterid': row.masterid,
                        'started_at': row.started_at,
                        'complete_at': row.complete_at,
                        'locks_duration_s': row.locks_duration_s,
                        'state_string': row.state_string,
                        'results': row.results,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.builds)
        return non_matched_rows

    def _thd_maybe_insert_build_properties(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildProperty)
        for row in matched_rows:
            conn.execute(
                self.model.build_properties.insert(),
                [
                    {
                        'buildid': row.buildid,
                        'name': row.name,
                        'value': json.dumps(row.value),
                        'source': row.source,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.build_properties)
        return non_matched_rows

    def _thd_maybe_insert_buildset(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Buildset)
        for row in matched_rows:
            conn.execute(
                self.model.buildsets.insert(),
                [
                    {
                        'id': row.id,
                        'external_idstring': row.external_idstring,
                        'reason': row.reason,
                        'submitted_at': row.submitted_at,
                        'complete': row_bool_to_int(row.complete),
                        'complete_at': row.complete_at,
                        'results': row.results,
                        'parent_buildid': None,
                        'parent_relationship': row.parent_relationship,
                        'rebuilt_buildid': None,
                    }
                ],
            )
        return rows  # filtered by _thd_maybe_insert_buildset_fk_columns

    def _thd_maybe_insert_buildset_fk_columns(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Buildset)
        for row in matched_rows:
            conn.execute(
                self.model.buildsets.update()
                .where(self.model.buildsets.c.id == row.id)
                .values(rebuilt_buildid=row.rebuilt_buildid, parent_buildid=row.parent_buildid)
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.buildsets)
        return non_matched_rows

    def _thd_maybe_insert_buildset_property(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildsetProperty)
        for row in matched_rows:
            conn.execute(
                self.model.buildset_properties.insert(),
                [
                    {
                        'buildsetid': row.buildsetid,
                        'property_name': row.property_name,
                        'property_value': row.property_value,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.buildset_properties)
        return non_matched_rows

    def _thd_maybe_insert_buildset_sourcestamp(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, BuildsetSourceStamp)
        for row in matched_rows:
            conn.execute(
                self.model.buildset_sourcestamps.insert(),
                [
                    {
                        'id': row.id,
                        'buildsetid': row.buildsetid,
                        'sourcestampid': row.sourcestampid,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.buildset_sourcestamps)
        return non_matched_rows

    def _thd_maybe_insert_change(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Change)
        for row in matched_rows:
            conn.execute(
                self.model.changes.insert(),
                [
                    {
                        'changeid': row.changeid,
                        'author': row.author,
                        'committer': row.committer,
                        'comments': row.comments,
                        'branch': row.branch,
                        'revision': row.revision,
                        'revlink': row.revlink,
                        'when_timestamp': row.when_timestamp,
                        'category': row.category,
                        'repository': row.repository,
                        'codebase': row.codebase,
                        'project': row.project,
                        'sourcestampid': row.sourcestampid,
                        'parent_changeids': row.parent_changeids,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.changes)
        return non_matched_rows

    def _thd_maybe_insert_change_file(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ChangeFile)
        for row in matched_rows:
            conn.execute(
                self.model.change_files.insert(),
                [{'changeid': row.changeid, 'filename': row.filename}],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.change_files)
        return non_matched_rows

    def _thd_maybe_insert_change_property(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ChangeProperty)
        for row in matched_rows:
            conn.execute(
                self.model.change_properties.insert(),
                [
                    {
                        'changeid': row.changeid,
                        'property_name': row.property_name,
                        'property_value': row.property_value,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.change_properties)
        return non_matched_rows

    def _thd_maybe_insert_change_user(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ChangeUser)
        for row in matched_rows:
            conn.execute(
                self.model.change_users.insert(),
                [{'changeid': row.changeid, 'uid': row.uid}],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.change_users)
        return non_matched_rows

    def _thd_maybe_insert_changesource(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ChangeSource)
        for row in matched_rows:
            conn.execute(
                self.model.changesources.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.changesources)
        return non_matched_rows

    def _thd_maybe_insert_changesource_master(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ChangeSourceMaster)
        for row in matched_rows:
            conn.execute(
                self.model.changesource_masters.insert(),
                [
                    {
                        'changesourceid': row.changesourceid,
                        'masterid': row.masterid,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.changesource_masters)
        return non_matched_rows

    def _thd_maybe_insert_log(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Log)
        for row in matched_rows:
            conn.execute(
                self.model.logs.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'slug': row.slug,
                        'stepid': row.stepid,
                        'complete': row.complete,
                        'num_lines': row.num_lines,
                        'type': row.type,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.logs)
        return non_matched_rows

    def _thd_maybe_insert_log_chunk(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, LogChunk)
        for row in matched_rows:
            conn.execute(
                self.model.logchunks.insert(),
                [
                    {
                        'logid': row.logid,
                        'first_line': row.first_line,
                        'last_line': row.last_line,
                        'content': row.content,
                        'compressed': row.compressed,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.logchunks)
        return non_matched_rows

    def _thd_maybe_insert_master(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Master)
        for row in matched_rows:
            conn.execute(
                self.model.masters.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                        'active': row_bool_to_int(row.active),
                        'last_active': int(row.last_active),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.masters)
        return non_matched_rows

    def _thd_maybe_insert_project(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Project)
        for row in matched_rows:
            conn.execute(
                self.model.projects.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                        'slug': row.slug,
                        'description': row.description,
                        'description_format': row.description_format,
                        'description_html': row.description_html,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.projects)
        return non_matched_rows

    def _thd_maybe_insert_codebase(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Codebase)
        for row in matched_rows:
            conn.execute(
                self.model.codebases.insert(),
                [
                    {
                        'id': row.id,
                        'projectid': row.projectid,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                        'slug': row.slug,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.codebases)
        return non_matched_rows

    def _thd_maybe_insert_codebase_commit(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, CodebaseCommit)
        for row in matched_rows:
            conn.execute(
                self.model.codebase_commits.insert(),
                [
                    {
                        'id': row.id,
                        'codebaseid': row.codebaseid,
                        'author': row.author,
                        'committer': row.committer,
                        'comments': row.comments,
                        'when_timestamp': int(row.when_timestamp),
                        'revision': row.revision,
                        'parent_commitid': row.parent_commitid,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.codebase_commits)
        return non_matched_rows

    def _thd_maybe_insert_codebase_branch(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, CodebaseBranch)
        for row in matched_rows:
            conn.execute(
                self.model.codebase_branches.insert(),
                [
                    {
                        'id': row.id,
                        'codebaseid': row.codebaseid,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                        'commitid': row.commitid,
                        'last_timestamp': int(row.last_timestamp),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.codebase_branches)
        return non_matched_rows

    def _thd_maybe_insert_scheduler_change(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, SchedulerChange)
        for row in matched_rows:
            conn.execute(
                self.model.scheduler_changes.insert(),
                [
                    {
                        'schedulerid': row.schedulerid,
                        'changeid': row.changeid,
                        'important': row.important,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.scheduler_changes)
        return non_matched_rows

    def _thd_maybe_insert_scheduler(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Scheduler)
        for row in matched_rows:
            conn.execute(
                self.model.schedulers.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                        'enabled': row.enabled,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.schedulers)
        return non_matched_rows

    def _thd_maybe_insert_scheduler_master(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, SchedulerMaster)
        for row in matched_rows:
            conn.execute(
                self.model.scheduler_masters.insert(),
                [
                    {
                        'schedulerid': row.schedulerid,
                        'masterid': row.masterid,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.scheduler_masters)
        return non_matched_rows

    def _thd_maybe_insert_patch(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Patch)
        for row in matched_rows:
            conn.execute(
                self.model.patches.insert(),
                [
                    {
                        'id': row.id,
                        'patchlevel': row.patchlevel,
                        'patch_base64': row.patch_base64,
                        'patch_author': row.patch_author,
                        'patch_comment': row.patch_comment,
                        'subdir': row.subdir,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.patches)
        return non_matched_rows

    def _thd_maybe_insert_sourcestamp(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, SourceStamp)
        for row in matched_rows:
            conn.execute(
                self.model.sourcestamps.insert(),
                [
                    {
                        'id': row.id,
                        'branch': row.branch,
                        'revision': row.revision,
                        'patchid': row.patchid,
                        'repository': row.repository,
                        'codebase': row.codebase,
                        'project': row.project,
                        'created_at': row.created_at,
                        'ss_hash': hash_columns(
                            row.branch,
                            row.revision,
                            row.repository,
                            row.project,
                            row.codebase,
                            row.patchid,
                        ),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.sourcestamps)
        return non_matched_rows

    def _thd_maybe_insert_object(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Object)
        for row in matched_rows:
            conn.execute(
                self.model.objects.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'class_name': row.class_name,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.objects)
        return non_matched_rows

    def _thd_maybe_insert_object_state(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ObjectState)
        for row in matched_rows:
            conn.execute(
                self.model.object_state.insert(),
                [
                    {
                        'objectid': row.objectid,
                        'name': row.name,
                        'value_json': row.value_json,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.object_state)
        return non_matched_rows

    def _thd_maybe_insert_step(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Step)
        for row in matched_rows:
            conn.execute(
                self.model.steps.insert(),
                [
                    {
                        'id': row.id,
                        'number': row.number,
                        'name': row.name,
                        'buildid': row.buildid,
                        'started_at': row.started_at,
                        'locks_acquired_at': row.locks_acquired_at,
                        'complete_at': row.complete_at,
                        'state_string': row.state_string,
                        'results': row.results,
                        'urls_json': row.urls_json,
                        'hidden': row_bool_to_int(row.hidden),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.steps)
        return non_matched_rows

    def _thd_maybe_insert_tag(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Tag)
        for row in matched_rows:
            conn.execute(
                self.model.tags.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'name_hash': hash_columns(row.name),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.tags)
        return non_matched_rows

    def _thd_maybe_insert_test_result_set(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, TestResultSet)
        for row in matched_rows:
            conn.execute(
                self.model.test_result_sets.insert(),
                [
                    {
                        'id': row.id,
                        'builderid': row.builderid,
                        'buildid': row.buildid,
                        'stepid': row.stepid,
                        'description': row.description,
                        'category': row.category,
                        'value_unit': row.value_unit,
                        'tests_passed': row.tests_passed,
                        'tests_failed': row.tests_failed,
                        'complete': row_bool_to_int(row.complete),
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.test_result_sets)
        return non_matched_rows

    def _thd_maybe_insert_test_name(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, TestName)
        for row in matched_rows:
            conn.execute(
                self.model.test_names.insert(),
                [
                    {
                        'id': row.id,
                        'builderid': row.builderid,
                        'name': row.name,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.test_names)
        return non_matched_rows

    def _thd_maybe_insert_test_code_path(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, TestCodePath)
        for row in matched_rows:
            conn.execute(
                self.model.test_code_paths.insert(),
                [
                    {
                        'id': row.id,
                        'builderid': row.builderid,
                        'path': row.path,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.test_code_paths)
        return non_matched_rows

    def _thd_maybe_insert_test_result(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, TestResult)
        for row in matched_rows:
            conn.execute(
                self.model.test_results.insert(),
                [
                    {
                        'id': row.id,
                        'builderid': row.builderid,
                        'test_result_setid': row.test_result_setid,
                        'test_nameid': row.test_nameid,
                        'test_code_pathid': row.test_code_pathid,
                        'line': row.line,
                        'duration_ns': row.duration_ns,
                        'value': row.value,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.test_results)
        return non_matched_rows

    def _thd_maybe_insert_user(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, User)
        for row in matched_rows:
            conn.execute(
                self.model.users.insert(),
                [
                    {
                        'uid': row.uid,
                        'identifier': row.identifier,
                        'bb_username': row.bb_username,
                        'bb_password': row.bb_password,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.users)
        return non_matched_rows

    def _thd_maybe_insert_user_info(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, UserInfo)
        for row in matched_rows:
            conn.execute(
                self.model.users_info.insert(),
                [
                    {
                        'uid': row.uid,
                        'attr_type': row.attr_type,
                        'attr_data': row.attr_data,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.users_info)
        return non_matched_rows

    def _thd_maybe_insert_worker(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, Worker)
        for row in matched_rows:
            conn.execute(
                self.model.workers.insert(),
                [
                    {
                        'id': row.id,
                        'name': row.name,
                        'info': row.info,
                        'paused': row.paused,
                        'pause_reason': row.pause_reason,
                        'graceful': row.graceful,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.workers)
        return non_matched_rows

    def _thd_maybe_insert_configured_worker(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ConfiguredWorker)
        for row in matched_rows:
            conn.execute(
                self.model.configured_workers.insert(),
                [
                    {
                        'id': row.id,
                        'buildermasterid': row.buildermasterid,
                        'workerid': row.workerid,
                    }
                ],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.configured_workers)
        return non_matched_rows

    def _thd_maybe_insert_connected_worker(self, conn, rows):
        matched_rows, non_matched_rows = self._match_rows(rows, ConnectedWorker)
        for row in matched_rows:
            conn.execute(
                self.model.connected_workers.insert(),
                [{'id': row.id, 'masterid': row.masterid, 'workerid': row.workerid}],
            )
        if matched_rows:
            self._thd_post_insert(conn, self.model.connected_workers)
        return non_matched_rows

    @defer.inlineCallbacks
    def insert_test_data(self, rows):
        """Insert a list of Row instances into the database"""

        def thd_insert_rows(conn):
            remaining = rows
            remaining = self._thd_maybe_insert_master(conn, remaining)
            remaining = self._thd_maybe_insert_project(conn, remaining)
            remaining = self._thd_maybe_insert_codebase(conn, remaining)
            remaining = self._thd_maybe_insert_codebase_commit(conn, remaining)
            remaining = self._thd_maybe_insert_codebase_branch(conn, remaining)
            remaining = self._thd_maybe_insert_builder(conn, remaining)
            remaining = self._thd_maybe_insert_tag(conn, remaining)
            remaining = self._thd_maybe_insert_worker(conn, remaining)
            remaining = self._thd_maybe_insert_user(conn, remaining)
            remaining = self._thd_maybe_insert_patch(conn, remaining)
            remaining = self._thd_maybe_insert_sourcestamp(conn, remaining)
            remaining = self._thd_maybe_insert_builder_master(conn, remaining)
            remaining = self._thd_maybe_insert_builder_tags(conn, remaining)
            remaining = self._thd_maybe_insert_buildset(conn, remaining)
            remaining = self._thd_maybe_insert_buildset_property(conn, remaining)
            remaining = self._thd_maybe_insert_buildset_sourcestamp(conn, remaining)
            remaining = self._thd_maybe_insert_buildrequest(conn, remaining)
            remaining = self._thd_maybe_insert_buildrequest_claim(conn, remaining)
            remaining = self._thd_maybe_insert_build(conn, remaining)
            remaining = self._thd_maybe_insert_build_data(conn, remaining)
            remaining = self._thd_maybe_insert_build_properties(conn, remaining)
            remaining = self._thd_maybe_insert_step(conn, remaining)
            remaining = self._thd_maybe_insert_change(conn, remaining)
            remaining = self._thd_maybe_insert_change_file(conn, remaining)
            remaining = self._thd_maybe_insert_change_property(conn, remaining)
            remaining = self._thd_maybe_insert_change_user(conn, remaining)
            remaining = self._thd_maybe_insert_changesource(conn, remaining)
            remaining = self._thd_maybe_insert_changesource_master(conn, remaining)
            remaining = self._thd_maybe_insert_log(conn, remaining)
            remaining = self._thd_maybe_insert_log_chunk(conn, remaining)
            remaining = self._thd_maybe_insert_scheduler(conn, remaining)
            remaining = self._thd_maybe_insert_scheduler_change(conn, remaining)
            remaining = self._thd_maybe_insert_scheduler_master(conn, remaining)
            remaining = self._thd_maybe_insert_object(conn, remaining)
            remaining = self._thd_maybe_insert_object_state(conn, remaining)
            remaining = self._thd_maybe_insert_test_result_set(conn, remaining)
            remaining = self._thd_maybe_insert_test_name(conn, remaining)
            remaining = self._thd_maybe_insert_test_code_path(conn, remaining)
            remaining = self._thd_maybe_insert_test_result(conn, remaining)
            remaining = self._thd_maybe_insert_user_info(conn, remaining)
            remaining = self._thd_maybe_insert_configured_worker(conn, remaining)
            remaining = self._thd_maybe_insert_connected_worker(conn, remaining)
            remaining = self._thd_maybe_insert_buildset_fk_columns(conn, remaining)

            self.testcase.assertEqual(remaining, [])

            conn.commit()

        yield self.pool.do(thd_insert_rows)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

import sqlalchemy as sa
from sqlalchemy.schema import MetaData
from twisted.python import log
from twisted.trial import unittest

from buildbot.db import model
from buildbot.util.sautils import withoutSqliteForeignKeys


def skip_for_dialect(dialect):
    """Decorator to skip a test for a particular SQLAlchemy dialect."""

    def dec(fn):
        def wrap(self, *args, **kwargs):
            if self.master.db._engine.dialect.name == dialect:
                raise unittest.SkipTest(f"Not supported on dialect '{dialect}'")
            return fn(self, *args, **kwargs)

        return wrap

    return dec


def get_trial_parallel_from_cwd(cwd):
    cwd = cwd.rstrip("/")
    last = os.path.basename(cwd)
    prev = os.path.basename(os.path.dirname(cwd))
    if last == "_trial_temp":
        return False
    if prev == "_trial_temp":
        try:
            return int(last)
        except ValueError:
            return None
    return None


def resolve_test_index_in_db_url(db_url: str) -> str:
    test_id = get_trial_parallel_from_cwd(os.getcwd())

    if "{TEST_ID}" in db_url:
        if test_id is None:
            raise RuntimeError("Database tests are run in parallel, but test index is unknown")

        return db_url.replace("{TEST_ID}", str(test_id or 0))

    if db_url == 'sqlite://':
        return db_url

    if test_id is not None and test_id is not False:
        if db_url.startswith('sqlite:///'):
            # Relative DB URLs in the test directory are fine.
            path = db_url[len('sqlite:///') :]
            if not os.path.relpath(path).startswith(".."):
                return db_url

        raise RuntimeError("Database tests cannnot run in parallel")

    return db_url


def resolve_test_db_url(db_url: str, sqlite_memory: bool) -> str:
    default_sqlite = 'sqlite://'
    if db_url is None:
        db_url = os.environ.get('BUILDBOT_TEST_DB_URL', default_sqlite)
        if not sqlite_memory and db_url == default_sqlite:
            db_url = "sqlite:///tmp.sqlite"

    return resolve_test_index_in_db_url(db_url)


def thd_clean_database(conn) -> None:
    # In general it's nearly impossible to do "bullet proof" database cleanup with SQLAlchemy
    # that will work on a range of databases and they configurations.
    #
    # Following approaches were considered.
    #
    # 1. Drop Buildbot Model schema:
    #
    #     model.Model.metadata.drop_all(bind=conn, checkfirst=True)
    #
    # Dropping schema from model is correct and working operation only if database schema is
    # exactly corresponds to the model schema.
    #
    # If it is not (e.g. migration script failed or migration results in old version of model),
    # then some tables outside model schema may be present, which may reference tables in the model
    # schema. In this case either dropping model schema will fail (if database enforces referential
    # integrity, e.g. PostgreSQL), or dropping left tables in the code below will fail (if database
    # allows removing of tables on which other tables have references, e.g. SQLite).
    #
    # 2. Introspect database contents and drop found tables.
    #
    #     meta = MetaData(bind=conn)
    #     meta.reflect()
    #     meta.drop_all()
    #
    # May fail if schema contains reference cycles (and Buildbot schema has them). Reflection looses
    # metadata about how reference cycles can be teared up (e.g. use_alter=True).
    # Introspection may fail if schema has invalid references (e.g. possible in SQLite).
    #
    # 3. What is actually needed here is accurate code for each engine and each engine configuration
    # that will drop all tables, indexes, constraints, etc in proper order or in a proper way
    # (using tables alternation, or DROP TABLE ... CASCADE, etc).
    #
    # Conclusion: use approach 2 with manually teared apart known reference cycles.

    try:
        meta = MetaData()

        # Reflect database contents. May fail, e.g. if table references
        # non-existent table in SQLite.
        meta.reflect(bind=conn)

        # Restore `use_alter` settings to break known reference cycles.
        # Main goal of this part is to remove SQLAlchemy warning
        # about reference cycle.

        # List of reference links (table_name, ref_table_name) that
        # should be broken by adding use_alter=True.
        table_referenced_table_links = [('buildsets', 'builds'), ('builds', 'buildrequests')]
        for table_name, ref_table_name in table_referenced_table_links:
            if table_name in meta.tables:
                table = meta.tables[table_name]
                for fkc in table.foreign_key_constraints:
                    if fkc.referred_table.name == ref_table_name:
                        fkc.use_alter = True

        # Drop all reflected tables and indices. May fail, e.g. if
        # SQLAlchemy wouldn't be able to break circular references.
        # Sqlalchemy fk support with sqlite is not yet perfect, so we must deactivate fk during
        # that operation, even though we made our possible to use use_alter
        with withoutSqliteForeignKeys(conn):
            meta.drop_all(bind=conn)
            conn.commit()

    except Exception:
        # sometimes this goes badly wrong; being able to see the schema
        # can be a big help
        if conn.engine.dialect.name == 'sqlite':
            r = conn.execute(sa.text("select sql from sqlite_master where type='table'"))
            log.msg("Current schema:")
            for row in r.fetchall():
                log.msg(row.sql)
        raise


def thd_create_tables(conn, table_names):
    table_names_set = set(table_names)
    tables = [t for t in model.Model.metadata.tables.values() if t.name in table_names_set]
    # Create tables using create_all() method. This way not only tables
    # and direct indices are created, but also deferred references
    # (that use use_alter=True in definition).
    model.Model.metadata.create_all(bind=conn, tables=tables, checkfirst=True)
    conn.commit()
